# lsqfitgp/_kernels/__init__.py
#
# Copyright (c) 2023, 2024, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

# Keep this file a pure import list.

from ._arma import (
    MA,
    AR,
)
from ._bart import BART
from ._basic import (
    Constant,
    White,
    ExpQuad,
    Linear,
    GammaExp,
    NNKernel,
    Gibbs,
    Periodic,
    Categorical,
    Rescaling,
    Expon,
    BagOfWords,
    HoleEffect,
    Cauchy,
    CausalExpQuad,
    Decaying,
    Log,
    Taylor,
)
from ._celerite import (
    Celerite,
    Harmonic,
)
from ._matern import (
    Maternp,
    Matern,
    Bessel,
)
from ._randomwalk import (
    Wiener,
    FracBrownian,
    WienerIntegral,
    OrnsteinUhlenbeck,
    BrownianBridge,
    StationaryFracBrownian,
)
from ._spectral import (
    Cos,
    Pink,
    Color,
    Sinc,
)
from ._wendland import (
    Wendland,
    Circular,
)
from ._zeta import Zeta

# TODO add explicit exponent parameter to all infinitely divisible kernels

# TODO new kernels LM(formula) and LMER(formula).



# lsqfitgp/_linalg/_toeplitz.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import jax
from jax import numpy as jnp
import numpy

from . import _seqalg

class SymSchur(_seqalg.Producer):
    """
    Cholesky decomposition of a symmetric Toeplitz matrix
    
    Adapted from TOEPLITZ_CHOLESKY by John Burkardt (LGPL license)
    http://people.sc.fsu.edu/~jburkardt/py_src/toeplitz_cholesky/toeplitz_cholesky.html
    """
    
    def __init__(self, t):
        """t = first row of the matrix"""
        t = jnp.asarray(t)
        assert t.ndim == 1
        # assert t[0] > 0, '1-th leading minor is not positive definite'
        self.t = t
    
    inputs = ()
    
    def init(self, n):
        t = self.t
        del self.t
        assert len(t) == n
        norm = t[0]
        t = t / norm
        self.g = jnp.stack([t, t])
        self.snorm = jnp.sqrt(norm)
    
    def iter_out(self, i):
        """i-th column of Cholesky factor L"""
        return self.g[0, :] * self.snorm
    
    def iter(self, i):
        g = self.g
        g = g.at[0, :].set(jnp.roll(g[0, :], 1))
        g = g.at[:, 0].set(0).at[:, i - 1].set(0)
        # assert g[0, i] > 0, 'what??'
        rho = -g[1, i] / g[0, i]
        # assert abs(rho) < 1, f'{i+1}-th leading minor is not positive definite'
        gamma = jnp.sqrt((1 - rho) * (1 + rho))
        self.g = (g + g[::-1] * rho) / gamma
    
    # TODO Schur, Levinson, and in general algorithms with triangular matrices
    # are not efficient as I implement them because due to the constraint of
    # fixed size in the jit I can not take advantage of the triangularity and
    # the number of operations doubles. This could be solved with a blocked
    # version which decreases the block size a few times, n blocks brings down
    # the increase factor to 1 + 1/n. The compilation time and code size would
    # be proportional to n. Aligning block size to powers of 2 would help
    # caching the compilation, bringing the compilation time to the normal one
    # after warmup (does jax compilation reuse functions or inline them?)
    #
    # Anyway, this 2 factor is not currently relevant.

class SymLevinson(_seqalg.Producer):
    """
    Cholesky decomposition of the *inverse* of a symmetric Toeplitz matrix
    
    Adapted form SuperGauss/DurbinLevinson.h (GPL license)
    https://cran.r-project.org/package=SuperGauss
    
    Note: Schur should be more accurate than Levinson
    """
    
    def __init__(self, t):
        """t = first row of the matrix"""
        t = jnp.asarray(t, float)
        assert t.ndim == 1
        # assert t[0] > 0, '1-th leading minor is not positive definite'
        self.t = t
    
    inputs = ()
        
    def init(self, n):
        self.phi1 = jnp.zeros(n)
        self.phi2 = jnp.zeros(n)
        self.nu = self.t[0]
        self.tlag = jnp.roll(self.t, -1)
        del self.t
        
    def iter_out(self, i):
        """i-th row of L^-1"""
        return -self.phi2.at[i].set(-1) / jnp.sqrt(self.nu)
        
    def iter(self, i):
        phi1 = self.phi1
        phi2 = self.phi2
        nu = self.nu
        tlag = self.tlag
        
        pi = i - 1
        rp = phi2 @ tlag
        phi1 = phi1.at[pi].set((tlag[pi] - rp) / nu)
        phi1 = phi1 - phi1[pi] * phi2
        # assert abs(phi1[pi]) < 1, f'{i+1}-th leading minor is not positive definite'
        nu = nu * (1 - phi1[pi]) * (1 + phi1[pi])
        phi2 = jnp.roll(phi1[::-1], i)
        
        self.phi1 = phi1
        self.phi2 = phi2
        self.nu = nu

@jax.jit
def chol(t):
    _, out = _seqalg.sequential_algorithm(len(t), [SymSchur(t), _seqalg.Stack(0)])
    return out.T

@jax.jit
def chol_solve(t, *bs):
    ops = [SymSchur(t)] + [_seqalg.SolveTriLowerColByFull(0, b) for b in bs]
    out = _seqalg.sequential_algorithm(len(t), ops)
    return out[1] if len(bs) == 1 else out[1:]

@jax.jit
def chol_matmul(t, b):
    ops = [SymSchur(t), _seqalg.Rows(b), _seqalg.MatMulColByRow(0, 1)]
    _, _, out = _seqalg.sequential_algorithm(len(t), ops)
    return out

@jax.jit
def chol_transp_matmul(t, b):
    ops = [SymSchur(t), _seqalg.MatMulRowByFull(0, b), _seqalg.Stack(1)]
    _, _, out = _seqalg.sequential_algorithm(len(t), ops)
    return out

@jax.jit
def logdet(t):
    _, out = _seqalg.sequential_algorithm(len(t), [SymSchur(t), _seqalg.SumLogDiag(0)])
    return 2 * out

@jax.jit
def solve(t, b):
    ops = [SymLevinson(t), _seqalg.MatMulRowByFull(0, b), _seqalg.MatMulColByRow(0, 1)]
    _, _, out = _seqalg.sequential_algorithm(len(t), ops)
    return out

@jax.jit
def chol_transp_solve(t, b):
    ops = [SymLevinson(t), _seqalg.Rows(b), _seqalg.MatMulColByRow(0, 1)]
    _, _, out = _seqalg.sequential_algorithm(len(t), ops)
    return out

def chol_solve_numpy(t, b, diageps=None):
    """
    
    Solve a linear system for the cholesky factor of a symmetric Toeplitz
    matrix. The algorithm is:
    
    t[0] += diageps
    m = toeplitz(t)
    l = chol(m)
    return solve(l, b)
    
    Numpy object arrays are supported. Broadcasts like matmul.
    
    Parameters
    ----------
    t : (..., n) array
        The first row or column of the matrix.
    b : (..., n, m) or (n,) array
        The right hand side of the linear system.
    diageps : scalar, optional
        Term added to the diagonal elements of the matrix for regularization.

    """
    
    t = numpy.array(t, subok=True)
    n = t.shape[-1]
        
    b = numpy.asanyarray(b)
    vec = b.ndim < 2
    if vec:
        b = b[:, None]
    assert b.shape[-2] == n
    
    t = t.astype(numpy.result_type(t, 0.1), copy=False)
    b = b.astype(numpy.result_type(b, 0.1), copy=False)
    
    if n == 0:
        shape = numpy.broadcast_shapes(t.shape[:-1], b.shape[:-2])
        shape += (n,) if vec else b.shape[-2:]
        dtype = numpy.result_type(t.dtype, b.dtype)
        return numpy.empty(shape, dtype)

    if diageps is not None:
        t[..., 0] += diageps

    if numpy.any(t[..., 0] <= 0):
        msg = '1-th leading minor is not positive definite'
        raise numpy.linalg.LinAlgError(msg)
    
    norm = numpy.copy(t[..., 0, None], subok=True)
    t /= norm
    invLb = numpy.copy(numpy.broadcast_arrays(b, t[..., None])[0], subok=True)
    prevLi = t
    g = numpy.stack([numpy.roll(t, 1, -1), t], -2)
    
    for i in range(1, n):
        
        assert numpy.all(g[..., 0, i] > 0)
        rho = -g[..., 1, i, None, None] / g[..., 0, i, None, None]

        if numpy.any(numpy.abs(rho) >= 1):
            msg = '{}-th leading minor is not positive definite'.format(i + 1)
            raise numpy.linalg.LinAlgError(msg)

        gamma = numpy.sqrt((1 - rho) * (1 + rho))
        g[..., :, i:] += g[..., ::-1, i:] * rho
        g[..., :, i:] /= gamma
        Li = g[..., 0, i:] # i-th column of L from row i
        invLb[..., i:, :] -= invLb[..., i - 1, None, :] * prevLi[..., i:, None]
        invLb[..., i, :] /= Li[..., 0, None]
        prevLi[..., i:] = Li
        g[..., 0, i:] = numpy.roll(g[..., 0, i:], 1, -1)

    invLb /= numpy.sqrt(norm[..., None])
    if vec:
        invLb = numpy.squeeze(invLb, -1)
    return invLb

def eigv_bound(t):
    """
    
    Bound the eigenvalues of a symmetric Toeplitz matrix.
    
    Parameters
    ----------
    t : array
        The first row of the matrix.
    
    Returns
    -------
    m : scalar
        Any eigenvalue `v` of the matrix satisfies `|v| <= m`.
    
    """
    s = jnp.abs(t)
    c = jnp.cumsum(s)
    d = c + c[::-1] - s[0]
    return jnp.max(d)



# lsqfitgp/_Kernel/_stationary.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from jax import numpy as jnp

from .. import _jaxext

from . import _util
from . import _crosskernel
from . import _kernel

class CrossStationaryKernel(_crosskernel.CrossKernel):
    """
    
    Subclass of `CrossKernel` for stationary kernels.

    A stationary kernel depends only on the difference, dimension by dimension,
    between its two arguments.

    Parameters
    ----------
    core : callable
        A function taking one positional argument ``delta = x - y`` and optional
        keyword arguments.
    input : {'signed', 'posabs', 'abs'}, default 'signed'
        If ``'signed'``, `kernel` is passed the bare difference. If
        ``'posabs'``, `kernel` is passed the absolute value of the difference,
        and the difference of equal points is a small number instead of zero. If
        ``'abs'``, the absolute value.
    **kw
        Additional keyword arguments are passed to the `CrossKernel`
        constructor.
            
    """

    def __new__(cls, core, *, input='signed', **kw):
        
        if input == 'posabs':
            dist = lambda x, y: _softabs(x - y)
        elif input == 'signed':
            dist = lambda x, y: x - y
        elif input == 'abs':
            dist = lambda x, y: jnp.abs(x - y)
        else:
            raise KeyError(input)
        
        def newcore(x, y, **kw):
            q = _util.ufunc_recurse_dtype(dist, x, y)
            return core(q, **kw)
        
        return super().__new__(cls, newcore, **kw)

    # TODO this class requires that, on both inputs, the thing depends only on
    # the distance. However, when transforming a stationary kernel, I often have
    # the property only on either side. Making a left/right class interferes
    # with _swap, because it is not clean to switch classes in the ancestors,
    # so there would be a class for a single side with a property.
    #
    # Maybe the elegant way to keep left/right properties separated is having
    # separate classes for the two processes.
    #
    # Drop the whole generic hierarchy. Keep a single class Kernel. Have two
    # attributes left and right which are instances of Process. Linops are
    # applied as methods on left and right, which operate on a partial
    # evaluation of the core. Kernel.__getattr__ sees if the missing attribute
    # is a method defined on both left and right and calls both of them; this
    # allows potential overrides for the joint transformation.
    #
    # The class logic operates on left/right. Linops looks one process at a
    # time. Algops are defined on the whole kernel but need somehow to indicate
    # how to change the classes of processes. Maybe a list of class-preserving
    # algops in the process.

class StationaryKernel(CrossStationaryKernel, _kernel.Kernel):
    pass

# make these transformations preserve the class StationaryKernel and upwards,
# other transformations are added by IsotropicKernel
StationaryKernel.inherit_transf('dim', intermediates=True)

def _eps(x):
    if jnp.issubdtype(x.dtype, jnp.inexact):
        return jnp.finfo(x.dtype).eps
        # finfo(x) does not work in numpy 1.20
    else:
        return jnp.finfo(jnp.empty(())).eps

def _softabs(x):
    return jnp.abs(x) + _eps(x)



# lsqfitgp/_kernels/_spectral.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from jax import numpy as jnp

from .. import _special
from .._Kernel import stationarykernel

@stationarykernel(derivable=True, maxdim=1)
def Cos(delta):
    """
    Cosine kernel.
    
    .. math::
        k(\\Delta) = \\cos(\\Delta)
        = \\cos x \\cos y + \\sin x \\sin y
    
    Samples from this kernel are harmonic functions. It can be multiplied with
    another kernel to introduce anticorrelations.
    
    """
    # TODO reference?
    return jnp.cos(delta)

    
@stationarykernel(maxdim=1, derivable=1, input='abs')
def Pink(delta, dw=1):
    """
    Pink noise kernel.
    
    .. math::
        k(\\Delta) &= \\frac 1 {\\log(1 + \\delta\\omega)}
        \\int_1^{1+\\delta\\omega} \\mathrm d\\omega
        \\frac{\\cos(\\omega\\Delta)}\\omega = \\\\
        &= \\frac {     \\operatorname{Ci}(\\Delta (1 + \\delta\\omega))
                        - \\operatorname{Ci}(\\Delta)                   }
        {\\log(1 + \\delta\\omega)}
    
    A process with power spectrum :math:`1/\\omega` truncated between 1 and
    :math:`1 + \\delta\\omega`. :math:`\\omega` is the angular frequency
    :math:`\\omega = 2\\pi f`. In the limit :math:`\\delta\\omega\\to\\infty`
    it becomes white noise. Derivable one time.
    
    """
    # TODO reference?

    l = _special.ci(delta)
    r = _special.ci(delta * (1 + dw))
    mean = delta * (1 + dw / 2)
    norm = jnp.log1p(dw)
    tol = jnp.sqrt(jnp.finfo(jnp.empty(0).dtype).eps)
    
    # TODO choose better this tolerance by estimating error terms
    
    return jnp.where(delta * dw < tol, jnp.cos(mean), (r - l) / norm)

def _color_derivable(n=2):
    return n // 2 - 1

@stationarykernel(maxdim=1, derivable=_color_derivable, input='abs')
def Color(delta, n=2):
    """
    Colored noise kernel.
    
    .. math::
        k(\\Delta) &= (n-1) \\Re E_n(-i\\Delta) = \\\\
        &= (n-1) \\int_1^\\infty \\mathrm d\\omega
        \\frac{\\cos(\\omega\\Delta)}{\\omega^n},
        \\quad n \\in \\mathbb N, n \\ge 2.
    
    A process with power spectrum :math:`1/\\omega^n` truncated below
    :math:`\\omega = 1`. :math:`\\omega` is the angular frequency
    :math:`\\omega = 2\\pi f`. Derivable :math:`\\lfloor n/2 \\rfloor - 1`
    times.
     
    """
    # TODO reference?
    
    # TODO real n > 1 => use usual series for small z and cont frac for large?
    
    # TODO integration limit dw like Pink, use None to mean inf since the
    # derivability changes and I can not use values for conditions due to the
    # jit, delete Pink
    
    # TODO The most general possible version I can imagine is int_a^b dw e^iwx
    # w^n with n in R. With n < -1 b=None is allowed, and a=None with n < 1,
    # meaning inf and 0. Since the general implementation would be slower, I can
    # detect n is an integer (at compile time) to use expn_imag. Afterward
    # delete Sinc and call this Power.
    #
    # This is the generalized incomplete gamma function with real argument and
    # imaginary parameters. DLMF says there is no known fixed-precision
    # implementation!
    #
    # int_l^h dw w^n e^-ixw =                               t = ixw
    #   = (ix)^-(n+1) int_ixl^ixh dt t^n e^-t =
    #   = (ix)^-(n+1) gammainc(n + 1, ixl, ixh) =           z = ix, a = n+1
    #   = z^-a gammainc(a, lz, hz) =
    #   = z^-a (gamma(a, hz) - gamma(a, lz)) =
    #   = z^-a (Gamma(a, lz) - Gamma(a, hz))
    #
    # int_0^1 dw w^n e^-ixw =
    #   = Gamma(n + 1) gamma*(n + 1, ix)
    # https://www.boost.org/doc/libs/1_79_0/libs/math/doc/html/math_toolkit/sf_gamma/igamma.html boost impl of gamma(a,x) and Gamma(a,x) for real x
    # https://dlmf.nist.gov/8.2.E7 gamma*(a, z)
    # https://dlmf.nist.gov/8.8.E14 d/da gamma*(a, z)
    # https://dlmf.nist.gov/8.8.E15 d^n/dz^n z^-a gamma(a, z)
    # https://dlmf.nist.gov/8.9 continued fractions for gamma* and Gamma
    # https://dlmf.nist.gov/8.11.E2 asymp series for Gamma(a, z)
    # https://dlmf.nist.gov/8.12 series for a->∞ in terms of erfc (do I have complex erfc? => yes scipy, no jax)
    # https://dlmf.nist.gov/8.19.E17 continued fraction for E_n(z)
    # https://dlmf.nist.gov/8.20.E6 series of E_n(z) for large n
    # https://dlmf.nist.gov/8.27.i reference to Padé of gammainc for complex z
    # 683.f impl of E_n(z) for complex z (license?) (amos1990)
    # https://specialfunctions.juliamath.org/stable/functions_list/#SpecialFunctions.expint implementation in julia for arbitrary complex arguments
    # https://github.com/JuliaMath/SpecialFunctions.jl/blob/36c547b4a270b6089b1baf7bec05707e9fb8c7f9/src/expint.jl#L446-L503
    
    # E_p(z)
    # following scipy's implementation, generalized to noninteger arguments:
    # if p > 50, use https://dlmf.nist.gov/8.20.ii (the dlmf reports it only
    #                for real z, need to follow the references)
    # elif |z| < 1 and p integer, use https://dlmf.nist.gov/8.19.E8
    #                p noninteger, use https://dlmf.nist.gov/8.19.E10 (need to
    #                                  sum the external term accurately)
    # else (|z| >= 1), use https://dlmf.nist.gov/8.19.E17
    
    # https://github.com/Radonirinaunimi/cmpx-spfunc incomplete gamma for
    # complex arguments (not well tested yet)
    
    # Parametrize with n = 1 + 2 nu like Matérn.
    
    # Bartosch, L. (2001). "Generation of colored noise". International Journal of Modern Physics C. 12 (6): 851–855. Bibcode:2001IJMPC..12..851B. doi:10.1142/S0129183101002012. S2CID 54500670.
    
    assert int(n) == n and n >= 2, n
    return (n - 1) * _special.expn_imag(n, delta).real
    
@stationarykernel(derivable=True, input='posabs', maxdim=1)
def Sinc(delta):
    """
    Sinc kernel.
    
    .. math:: k(\\Delta) = \\operatorname{sinc}(\\Delta) =
        \\frac{\\sin(\\pi\\Delta)}{\\pi\\Delta}.
    
    Reference: Tobar (2019).
    """
    return _special.sinc(delta)
    
    # TODO is this isotropic? My current guess is that it works up to some
    # dimension due to a coincidence but is not in general isotropic.



# lsqfitgp/_jaxext/_fasthash.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

# JAX port of fasthash https://github.com/ztanml/fast-hash, original license:

# The MIT License
#
# Copyright (C) 2012 Zilong Tan (eric.zltan@gmail.com)
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import functools

from jax import numpy as jnp
from jax import lax
import jax

# TODO this breaks down if jax_enable_x64=False. I have to write the operations
# on 64 bit integers in terms of operations on 32 bit integers, and always
# manipulate arrays with 2x uint32. I bet I can find such translations on the
# internet. I can make an interface for something that switches automatically
# between the two modes.

# TODO move to _special

# Compression function for Merkle-Damgard construction.
# This function is generated using the framework provided.
def mix(h):
    h ^= h >> 23
    h *= jnp.array(0x2127599bf4325c37, jnp.uint64)
    h ^= h >> 47
    return h

@functools.partial(jax.jit, static_argnames=('unroll',))
def fasthash64(buf, seed, *, unroll=4):
    # buf = jnp.asarray(buf) # needed without jit
    seed = jnp.array(seed, jnp.uint64)
    assert seed.dtype == jnp.uint64 # check that jax_enable_x64=True
    assert buf.ndim == 1
    buf = buf.view(jnp.uint8)
    m = jnp.array(0x880355f21e6d1965, jnp.uint64)
    pos = buf[:buf.size - buf.size % 8].view(jnp.uint64)
    h = seed ^ (buf.size * m)

    def loop(carry, v):
        h, m = carry
        h ^= mix(v)
        h *= m
        return (h, m), None
    (h, _), _ = lax.scan(loop, (h, m), pos, unroll=unroll)

    pos2 = buf[pos.nbytes:]
    assert pos2.size < 8
    assert pos.nbytes + pos2.size == buf.size
    v = jnp.array(0, jnp.uint64)

    if pos2.size >= 7: v ^= pos2[6].astype(jnp.uint64) << 48
    if pos2.size >= 6: v ^= pos2[5].astype(jnp.uint64) << 40
    if pos2.size >= 5: v ^= pos2[4].astype(jnp.uint64) << 32
    if pos2.size >= 4: v ^= pos2[3].astype(jnp.uint64) << 24
    if pos2.size >= 3: v ^= pos2[2].astype(jnp.uint64) << 16
    if pos2.size >= 2: v ^= pos2[1].astype(jnp.uint64) << 8
    if pos2.size >= 1: v ^= pos2[0].astype(jnp.uint64)
    if pos2.size:
        h ^= mix(v)
        h *= m

    assert h.dtype == jnp.uint64
    return mix(h)

def fasthash32(buf, seed):
    seed = jnp.array(seed, jnp.uint32)
    h = fasthash64(buf, seed)
    return (h - (h >> 32)).astype(jnp.uint32)



# lsqfitgp/_special/_gamma.py
#
# Copyright (c) 2022, 2023, 2024, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from jax import numpy as jnp
from jax.scipy import special as jspecial

from .. import _jaxext

def sgngamma(x):
    return jnp.where((x > 0) | (x % 2 < 1), 1, -1)

def gamma(x):
    return sgngamma(x) * jnp.exp(jspecial.gammaln(x))

def poch(x, k):
    return jnp.exp(jspecial.gammaln(x + k) - jspecial.gammaln(x)) # DLMF 5.2.5
    # TODO does not handle properly special cases with x and/or k nonpositive
    # integers

def gamma_incr(x, e):
    """
    Compute Γ(x+e) / (Γ(x)Γ(1+e)) - 1 accurately for x >= 2 and |e| < 1/2
    """
    
    # G(x + e) / G(x)G(1+e) - 1 =
    # = expm1(log(G(x + e) / G(x)G(1+e))) =
    # = expm1(log G(x + e) - log G(x) - log G(1 + e))
    
    t = _jaxext.float_type(x, e)
    n = 23 if t == jnp.float64 else 10
    # n such that 1/2^n 1/n! d^n/dx^n log G(x) |_x=2 < eps
    k = jnp.arange(n).reshape((n,) + (1,) * max(x.ndim, e.ndim))
    coef = jspecial.polygamma(k, x)
    fact = jnp.cumprod(1 + k, 0, t)
    coef /= fact
    gammaln = e * jnp.polyval(coef[::-1], e)
    return jnp.expm1(gammaln - gammaln1(e))
    
    # Like gammaln1, I thought that writing as log Γ(1+x+e) - log Γ(1+x) +
    # - log1p(e/x) would increase accuracy, instead it deteriorates

def gammaln1(x):
    """ compute log Γ(1+x) accurately for |x| <= 1/2 """
    
    t = _jaxext.float_type(x)
    coef = jnp.array(_gammaln1_coef_1[:48], t) # 48 found by trial and error
    return x * jnp.polyval(coef[::-1], x)
    
    # I thought that writing this as log Γ(2+x) - log1p(x) would be more
    # accurate but it isn't, probably there's a cancellation for some values of
    # x

_gammaln1_coef_1 = [        # = _gen_gammaln1_coef(53, 1)
    -0.5772156649015329,
    0.8224670334241132,
    -0.40068563438653143,
    0.27058080842778454,
    -0.20738555102867398,
    0.1695571769974082,
    -0.1440498967688461,
    0.12550966952474304,
    -0.11133426586956469,
    0.1000994575127818,
    -0.09095401714582904,
    0.083353840546109,
    -0.0769325164113522,
    0.07143294629536133,
    -0.06666870588242046,
    0.06250095514121304,
    -0.058823978658684585,
    0.055555767627403614,
    -0.05263167937961666,
    0.05000004769810169,
    -0.047619070330142226,
    0.04545455629320467,
    -0.04347826605304026,
    0.04166666915034121,
    -0.04000000119214014,
    0.03846153903467518,
    -0.037037037312989324,
    0.035714285847333355,
    -0.034482758684919304,
    0.03333333336437758,
    -0.03225806453115042,
    0.03125000000727597,
    -0.030303030306558044,
    0.029411764707594344,
    -0.02857142857226011,
    0.027777777778181998,
    -0.027027027027223673,
    0.02631578947377995,
    -0.025641025641072283,
    0.025000000000022737,
    -0.024390243902450117,
    0.023809523809529224,
    -0.023255813953491015,
    0.02272727272727402,
    -0.022222222222222855,
    0.021739130434782917,
    -0.021276595744681003,
    0.02083333333333341,
    -0.02040816326530616,
    0.020000000000000018,
    -0.019607843137254912,
    0.019230769230769235,
    -0.01886792452830189,
]

def _gen_gammaln1_coef(n, x): # pragma: no cover
    """ compute Taylor coefficients of log Γ(x) """
    import mpmath as mp
    with mp.workdps(32):
        return [float(mp.polygamma(k, x) / mp.fac(k + 1)) for k in range(n)]



# lsqfitgp/_Kernel/_util.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import numbers
import operator

import numpy
import jax
from jax import numpy as jnp
from jax import tree_util

from .. import _array

def is_numerical_scalar(x):
    return (
        isinstance(x, numbers.Number) or
        (isinstance(x, (numpy.ndarray, jnp.ndarray)) and x.ndim == 0)
    )
    # do not use jnp.isscalar because it returns False for strongly
    # typed 0-dim arrays; do not use jnp.ndim(•) == 0 because it accepts
    # non-numerical types

def is_nonnegative_integer_scalar(x):
    if isinstance(x, numbers.Integral) and x >= 0:
        # python scalars and numpy scalars
        return True
    if isinstance(x, numpy.ndarray) and x.ndim == 0 and numpy.issubdtype(x.dtype, numpy.integer) and x.item() >= 0:
        # 0-dim numpy arrays
        return True
    if isinstance(x, jnp.ndarray) and x.ndim == 0 and jnp.issubdtype(x.dtype, jnp.integer):
        try:
            # concrete jax arrays
            return x.item() >= 0
        except jax.errors.ConcretizationTypeError:
            # jax tracers
            return jnp.issubdtype(x.dtype, jnp.unsignedinteger)
    return False

def is_scalar_cond_trueontracer(x, cond):
    if isinstance(x, numbers.Number) and cond(x):
        # python scalars and numpy scalars
        return True
    if isinstance(x, numpy.ndarray) and x.ndim == 0 and numpy.issubdtype(x.dtype, numpy.number) and cond(x.item()):
        # 0-dim numpy arrays
        return True
    if isinstance(x, jnp.ndarray) and x.ndim == 0 and jnp.issubdtype(x.dtype, jnp.number):
        try:
            # concrete jax arrays
            return cond(x.item())
        except jax.errors.ConcretizationTypeError:
            # jax tracers
            return True
    return False

def is_nonnegative_scalar_trueontracer(x):
    return is_scalar_cond_trueontracer(x, lambda x: x >= 0)

# TODO reimplement with tree_reduce, closuring ndim to recognize shaped fields
def _reduce_recurse_dtype(fun, args, reductor, npreductor, jnpreductor, **kw):
    x = args[0]
    if x.dtype.names is None:
        return fun(*args, **kw)
    else:
        acc = None
        for name in x.dtype.names:
            recargs = tuple(arg[name] for arg in args)
            result = _reduce_recurse_dtype(fun, recargs, reductor, npreductor, jnpreductor, **kw)
            
            dtype = x.dtype[name]
            if dtype.ndim:
                axis = tuple(range(-dtype.ndim, 0))
                red = jnpreductor if isinstance(result, jnp.ndarray) else npreductor
                result = red(result, axis=axis)
            
            if acc is None:
                acc = result
            else:
                acc = reductor(acc, result)
        
        assert acc.shape == _array.broadcast(*args).shape
        return acc

def sum_recurse_dtype(fun, *args, **kw):
    return _reduce_recurse_dtype(fun, args, operator.add, numpy.sum, jnp.sum, **kw)

def prod_recurse_dtype(fun, *args, **kw):
    return _reduce_recurse_dtype(fun, args, operator.mul, numpy.prod, jnp.prod, **kw)

def ufunc_recurse_dtype(ufunc, x, *args):
    """ apply an ufunc to all the leaf fields """
    
    allargs = (x, *args)
    expected_shape = jnp.broadcast_shapes(*(x.shape for x in allargs))

    if x.dtype.names is None:
        out = ufunc(*allargs)
    else:
        args = map(_array.StructuredArray, allargs)
        out = tree_util.tree_map(ufunc, *args)
    
    assert out.shape == expected_shape
    return out



# lsqfitgp/_special/_sinc.py
#
# Copyright (c) 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from jax import numpy as jnp
from jax.scipy import special as jspecial

from . import _taylor

def coefgen_sinc(s, e):
    m = jnp.arange(s, e)
    return (-1) ** m / jnp.exp(jspecial.gammaln(2 + 2 * m))

def sinc(x):
    nearzero = _taylor.taylor(coefgen_sinc, (), 0, 6, jnp.square(jnp.pi * x))
    normal = jnp.sinc(x)
    return jnp.where(jnp.abs(x) < 1e-1, nearzero, normal)



# lsqfitgp/_GP/_processes.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import abc
import numbers

import numpy
from jax import numpy as jnp

from .. import _Kernel
from .. import _Deriv

from . import _base

class GPProcesses(_base.GPBase):

    def __init__(self, *, covfun):
        self._procs = {} # proc key -> _Proc
        self._kernels = {} # (proc key, proc key) -> CrossKernel
        if covfun is not None:
            if not isinstance(covfun, _Kernel.Kernel):
                raise TypeError('covariance function must be of class Kernel')
            self._procs[self.DefaultProcess] = self._ProcKernel(covfun)

    def _clone(self):
        newself = super()._clone()
        newself._procs = self._procs.copy()
        newself._kernels = self._kernels.copy()
        return newself

    class _Proc(abc.ABC):
        """
        Abstract base class for an object holding information about a process
        in a GP object.
        """
    
        @abc.abstractmethod
        def __init__(self): # pragma: no cover
            pass
    
    class _ProcKernel(_Proc):
        """An independent process defined with a kernel"""
    
        def __init__(self, kernel, deriv=0):
            assert isinstance(kernel, _Kernel.Kernel)
            self.kernel = kernel
            self.deriv = deriv
    
    class _ProcTransf(_Proc):
        """A process defined as a linear transformation of other processes"""
    
        def __init__(self, ops, deriv):
            """ops = dict proc key -> callable"""
            self.ops = ops
            self.deriv = deriv

    class _ProcLinTransf(_Proc):
    
        def __init__(self, transf, keys, deriv):
            self.transf = transf
            self.keys = keys
            self.deriv = deriv

    class _ProcKernelTransf(_Proc):
        """A process defined by an operation on the kernel of another process"""
    
        def __init__(self, proc, transfname, arg):
            """proc = proc key, transfname = Kernel transfname, arg = argument to transf """
            self.proc = proc
            self.transfname = transfname
            self.arg = arg
        
    _zerokernel = _Kernel.Zero()

    @_base.newself
    def defproc(self, key, kernel=None, *, deriv=0):
        """
        
        Add an independent process.
        
        Parameters
        ----------
        key : hashable
            The name that identifies the process in the GP object.
        kernel : Kernel
            A kernel for the process. If None, use the default kernel. The
            difference between the default process and a process defined with
            the default kernel is that, although they have the same kernel,
            they are independent.
        deriv : Deriv-like
            Derivatives to take on the process defined by the kernel.
                
        """
        
        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
        
        if kernel is None:
            kernel = self._procs[self.DefaultProcess].kernel
        
        deriv = _Deriv.Deriv(deriv)
        
        self._procs[key] = self._ProcKernel(kernel, deriv)
    
    @_base.newself
    def deftransf(self, key, ops, *, deriv=0):
        """
        
        Define a new process as a linear combination of other processes.
        
        Let f_i(x), i = 1, 2, ... be already defined processes, and g_i(x) be
        deterministic functions. The new process is defined as
        
            h(x) = g_1(x) f_1(x) + g_2(x) f_2(x) + ...
        
        Parameters
        ----------
        key : hashable
            The name that identifies the new process in the GP object.
        ops : dict
            A dictionary mapping process keys to scalars or scalar
            functions. The functions must take an argument of the same kind
            of the domain of the process.
        deriv : Deriv-like
            The linear combination is derived as specified by this
            parameter.
        
        """
        
        for k, func in ops.items():
            if k not in self._procs:
                raise KeyError(f'process key {k!r} not in GP object')
            if not _Kernel.is_numerical_scalar(func) and not callable(func):
                raise TypeError(f'object of type {type(func)!r} for key {k!r} is neither scalar nor callable')

        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
        
        deriv = _Deriv.Deriv(deriv)
        
        self._procs[key] = self._ProcTransf(ops, deriv)
        
        # we could implement deftransf in terms of deflintransf with
        # the following code, but deftransf has linear kernel building
        # cost so I'm leaving it around (probably not significant anyway)
        # functions = [
        #     op if callable(op)
        #     else (lambda x: lambda _: x)(op)
        #     for op in ops.values()
        # ]
        # def equivalent_lintransf(*procs):
        #     def fun(x):
        #         out = None
        #         for fun, proc in zip(functions, procs):
        #             this = fun(x) * proc(x)
        #             out = this if out is None else out + this
        #         return out
        #     return fun
        # self.deflintransf(key, equivalent_lintransf, list(ops.keys()), deriv=deriv, checklin=False)
    
    @_base.newself
    def deflintransf(self, key, transf, procs, *, deriv=0, checklin=False):
        """
        
        Define a new process as a linear combination of other processes.
        
        Let f_i(x), i = 1, 2, ... be already defined processes, and T
        a linear map from processes to a single process. The new process is
        
            h(x) = T(f_1, f_2, ...)(x).
        
        Parameters
        ----------
        key : hashable
            The name that identifies the new process in the GP object.
        transf : callable
            A function with signature ``transf(callable, callable, ...) -> callable``.
        procs : sequence
            The keys of the processes to be passed to the transformation.
        deriv : Deriv-like
            The linear combination is derived as specified by this
            parameter.
        checklin : bool
            If True, check if the transformation is linear. Default False.
        
        Notes
        -----
        The linearity check may fail if the transformation does nontrivial
        operations with the inner function input.
        
        """
        
        # TODO support procs being a single key
        
        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
    
        for k in procs:
            if k not in self._procs:
                raise KeyError(k)
        
        deriv = _Deriv.Deriv(deriv)
        
        if len(procs) == 0:
            self._procs[key] = self._ProcKernel(self._zerokernel)
            return
        
        if checklin is None:
            checklin = self._checklin
        if checklin:
            mockup_function = lambda a: lambda _: a
            # TODO this array mockup fails with jax functions
            class Mockup(numpy.ndarray):
                __getitem__ = lambda *_: Mockup((0,))
                __getattr__ = __getitem__
            def checktransf(*arrays):
                functions = [mockup_function(a) for a in arrays]
                return transf(*functions)(Mockup((0,)))
            shapes = [(11,)] * len(procs)
            self._checklinear(checktransf, shapes, elementwise=True)
        
        self._procs[key] = self._ProcLinTransf(transf, procs, deriv)

    @_base.newself
    def deflinop(self, key, transfname, arg, proc):
        """
        
        Define a new process as the transformation of an existing one.
        
        Parameters
        ----------
        key : hashable
            Key for the new process.
        transfname : hashable
            A transformation recognized by the `~CrossKernel.transf` method
            of the kernel.
        arg :
            A valid argument to the transformation.
        proc : hashable
            Key of the process to be transformed.
        
        """
        
        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
        if proc not in self._procs:
            raise KeyError(f'process {proc!r} not found')
        self._procs[key] = self._ProcKernelTransf(proc, transfname, arg)
    
    def defderiv(self, key, deriv, proc):
        """
        
        Define a new process as the derivative of an existing one.
        
        .. math::
            g(x) = \\frac{\\partial^n}{\\partial x^n} f(x)
        
        Parameters
        ----------
        key : hashable
            The key of the new process.
        deriv : Deriv-like
            Derivation order.
        proc : hashable
            The key of the process to be derived.
        
        Returns
        -------
        gp : GP
            A new GP object with the applied modifications.

        """
        deriv = _Deriv.Deriv(deriv)
        return self.deflinop(key, 'diff', deriv, proc)
    
    def defxtransf(self, key, transf, proc):
        """
        
        Define a new process by transforming the inputs of another one.
        
        .. math::
            g(x) = f(T(x))
        
        Parameters
        ----------
        key : hashable
            The key of the new process.
        transf : callable
            A function mapping the new kind input to the input expected by the
            transformed process.
        proc : hashable
            The key of the process to be transformed.
        
        Returns
        -------
        gp : GP
            A new GP object with the applied modifications.

        """
        assert callable(transf)
        return self.deflinop(key, 'xtransf', transf, proc)
    
    def defrescale(self, key, scalefun, proc):
        """
        
        Define a new process as a rescaling of an existing one.
        
        .. math::
            g(x) = s(x)f(x)
        
        Parameters
        ----------
        key : hashable
            The key of the new process.
        scalefun : callable
            A function from the domain of the process to a scalar.
        proc : hashable
            The key of the process to be transformed.
        
        Returns
        -------
        gp : GP
            A new GP object with the applied modifications.

        """
        assert callable(scalefun)
        return self.deflinop(key, 'rescale', scalefun, proc)
    
    def _crosskernel(self, xpkey, ypkey):
        
        # Check if the kernel is in cache.
        cache = self._kernels.get((xpkey, ypkey))
        if cache is not None:
            return cache
        
        # Compute the kernel.
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        if isinstance(xp, self._ProcKernel) and isinstance(yp, self._ProcKernel):
            kernel = self._crosskernel_kernels(xpkey, ypkey)
        elif isinstance(xp, self._ProcTransf):
            kernel = self._crosskernel_transf_any(xpkey, ypkey)
        elif isinstance(yp, self._ProcTransf):
            kernel = self._crosskernel_transf_any(ypkey, xpkey)._swap()
        elif isinstance(xp, self._ProcLinTransf):
            kernel = self._crosskernel_lintransf_any(xpkey, ypkey)
        elif isinstance(yp, self._ProcLinTransf):
            kernel = self._crosskernel_lintransf_any(ypkey, xpkey)._swap()
        elif isinstance(xp, self._ProcKernelTransf):
            kernel = self._crosskernel_kerneltransf_any(xpkey, ypkey)
        elif isinstance(yp, self._ProcKernelTransf):
            kernel = self._crosskernel_kerneltransf_any(ypkey, xpkey)._swap()
        else: # pragma: no cover
            raise TypeError(f'unrecognized process types {type(xp)!r} and {type(yp)!r}')
        
        # Save cache.
        self._kernels[xpkey, ypkey] = kernel
        self._kernels[ypkey, xpkey] = kernel._swap()
        
        return kernel
        
    def _crosskernel_kernels(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        if xp is yp:
            return xp.kernel.linop('diff', xp.deriv, xp.deriv)
        else:
            return self._zerokernel
    
    def _crosskernel_transf_any(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        kernelsum = self._zerokernel
        
        for pkey, factor in xp.ops.items():
            kernel = self._crosskernel(pkey, ypkey)
            if kernel is self._zerokernel:
                continue
            
            if not callable(factor):
                factor = (lambda f: lambda _: f)(factor)
            kernel = kernel.linop('rescale', factor, None)
            
            if kernelsum is self._zerokernel:
                kernelsum = kernel
            else:
                kernelsum += kernel
                
        return kernelsum.linop('diff', xp.deriv, 0)
    
    def _crosskernel_lintransf_any(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        kernels = [self._crosskernel(pk, ypkey) for pk in xp.keys]
        kernel = _Kernel.CrossKernel._nary(xp.transf, kernels, _Kernel.CrossKernel._side.LEFT)
        kernel = kernel.linop('diff', xp.deriv, 0)
        
        return kernel
    
    def _crosskernel_kerneltransf_any(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        if xp is yp:
            basekernel = self._crosskernel(xp.proc, xp.proc)
            # I could avoid handling this case separately but it allows to
            # skip defining two-step transformations A -> CrossAT -> T
        else:
            basekernel = self._crosskernel(xp.proc, ypkey)
        
        if basekernel is self._zerokernel:
            return self._zerokernel
        elif xp is yp:
            return basekernel.linop(xp.transfname, xp.arg)
        else:
            return basekernel.linop(xp.transfname, xp.arg, None)



# lsqfitgp/_GP/_gp.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from . import _base, _compute, _elements, _processes

class GP(_compute.GPCompute, _elements.GPElements, _processes.GPProcesses):
    """
    
    Object that represents a Gaussian process.

    A `GP` is structured like a pair of dictionaries, one for "processes", and
    one for "elements". The processes represent independent Gaussian processes,
    i.e., infinite-dimensional Normally distributed variables. The elements
    represent finite-dimensional Normal variables, typically finite subsets of
    the processes.

    The methods to define processes start with "def", while those to define
    elements starts with "add". The basic methods are `defproc` and `addx`.
    
    A `GP` object is immutable. Methods that modify the Gaussian process return
    a new object which differs only in the requested modification, leaving the
    original untouched.

    Parameters
    ----------
    covfun : Kernel, optional
        An instance of `Kernel` representing the covariance kernel of the
        default process of the GP object. It can be left unspecified.
    solver : str, default 'chol'
        The algorithm used to decompose the prior covariance matrix. See
        `decompose` for the available solvers.
    checkpos : bool, default True
        Raise a `LinAlgError` if the prior covariance matrix turns out non
        positive within numerical error.
    checksym : bool, default True
        Check that the prior covariance matrix is symmetric.
    checkfinite : bool, default True
        Check that the prior covariance matrix does not contain infs or nans.
    checklin : bool, default True
        The method `addlintransf` will check that the given transformation is
        linear on a random input tensor.
    posepsfac : number, default 1
        The threshold used to check if the prior covariance matrix is positive
        definite is multiplied by this factor.
    halfmatrix : bool, default False
        If ``checksym=False``, compute only half of the covariance matrices by
        unrolling their lower triangular part as flat arrays. This may actually
        be a large performance hit if the input arrays have large item size or
        if the implementation of the kernel takes advantage of non-broadcasted
        inputs.
    **kw
        Additional keyword arguments are passed to the solver, see `decompose`.

    Methods
    -------
    addx
        Add points where a process is evaluated.
    addlintransf
        Define a finite linear transformation.
    addtransf
        Define a finite linear transformation with explicit coefficients.
    addcov
        Introduce a set of user-provided prior covariance matrix blocks.
    defproc
        Define a new independent process with a kernel.
    deflintransf
        Define a pointwise linear transformation.
    deftransf
        Define a pointwise linear transformation with explicit coefficients.
    deflinop
        Define an arbitrary linear transformation through a kernel method.
    defderiv
        Define a process as the derivative of another one.
    defxtransf
        Define a process with transformed inputs.
    defrescale
        Rescale a process.
    prior
        Compute the prior.
    pred
        Compute the posterior.
    predfromfit
        Like `pred` with ``fromdata=False``.
    predfromdata
        Like `pred` with ``fromdata=True``.
    marginal_likelihood
        Compute the probability density.
    decompose
        Decompose a pos. semidef. matrix.

    Attributes
    ----------
    DefaultProcess :
        Key that identifies the default process.
    
    """

    def __init__(self,
        covfun=None,
        *,
        solver='chol',
        checkpos=True,
        checksym=True,
        checkfinite=True,
        checklin=True,
        posepsfac=1,
        halfmatrix=False,
        **kw,
    ):
        _base.GPBase.__init__(self, checkfinite=checkfinite, checklin=checklin)
        _processes.GPProcesses.__init__(self, covfun=covfun)
        _elements.GPElements.__init__(self, checkpos=checkpos, checksym=checksym, posepsfac=posepsfac, halfmatrix=halfmatrix)
        _compute.GPCompute.__init__(self, solver=solver, solverkw=kw)



# lsqfitgp/_fit.py
#
# Copyright (c) 2020, 2022, 2023, 2024, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import re
import warnings
import functools
import time
import textwrap
import datetime

import gvar
import jax
from jax import numpy as jnp
import numpy
from scipy import optimize
from jax import tree_util

from . import _GP
from . import _linalg
from . import _jaxext
from . import _gvarext
from . import _array

# TODO the following token_ thing functionality may be provided by jax in the
# future, follow the developments

@functools.singledispatch
def token_getter(x):
    return x

@functools.singledispatch
def token_setter(x, token):
    return token

@token_getter.register(jnp.ndarray)
@token_getter.register(numpy.ndarray)
def _(x):
    return x[x.ndim * (0,)] if x.size else x

@token_setter.register(jnp.ndarray)
@token_setter.register(numpy.ndarray)
def _(x, token):
    x = jnp.asarray(x)
    return x.at[x.ndim * (0,)].set(token) if x.size else token

def token_map_leaf(func, x):
    if isinstance(x, (jnp.ndarray, numpy.ndarray)):
        token = token_getter(x)
        @jax.custom_jvp
        def jaxfunc(token):
            return jax.pure_callback(func, token, token, vectorized=True)
        @jaxfunc.defjvp
        def _(p, t):
            return (jaxfunc(*p), *t)
        token = jaxfunc(token)
        return token_setter(x, token)
    else:
        token = token_getter(x)
        token = func(token)
        return token_setter(x, token)

def token_map(func, x):
    return tree_util.tree_map(lambda x: token_map_leaf(func, x), x)

class Logger:
    """ Class to manage a log. Can be used as superclass. Each line of the log
    has a verbosity level (an integer >= 0) and is printed only if this level is
    below a threshold. All lines are saved and the log can be retrieved. """

    def __init__(self, target_verbosity=0):
        """ set the threshold used to exclude log lines """
        self._verbosity = target_verbosity
        self._loggedlines = []
    
    def _indent(self, text, level=0):
        """ indent a text by provided level or by global current level """
        level = max(0, level + self.loglevel._level)
        prefix = 4 * level * ' '
        return textwrap.indent(text, prefix)

    def _select(self, verbosity, target_verbosity=None):
        if target_verbosity is None:
            target_verbosity = self._verbosity
        if isinstance(verbosity, int):
            return target_verbosity >= verbosity
        else:
            return target_verbosity in verbosity
    
    def log(self, message, verbosity=1, *, level=0):
        """
        Print and record a message.

        Parameters
        ----------
        message : str
            The message to print. A newline is added unconditionally.
        verbosity : int or set, default 1
            The verbosity level(s) at which the message is printed. If an
            integer, it's printed at all levels >= that integer. If a set, at
            the specified levels.
        level : int, default 0
            The indentation level of the message.
        """
        if self._select(verbosity):
            print(self._indent(message, level))
        self._loggedlines.append((message, verbosity, level + self.loglevel._level))

    def getlog(self, target_verbosity=None, *, base_level=0):
        """ return all logged line as a single string """
        return '\n'.join(
            self._indent(message, base_level + level)
            for message, verbosity, level in self._loggedlines
            if self._select(verbosity, target_verbosity)
        )

    class _LogLevel:
        """ shared context manager to indent messages """
        
        _level = 0
                
        @classmethod
        def __enter__(cls):
            cls._level += 1
        
        @classmethod
        def __exit__(cls, *_):
            cls._level -= 1
    
    loglevel = _LogLevel()

class empbayes_fit(Logger):

    SEPARATE_JAC = False

    def __init__(
        self,
        hyperprior,
        gpfactory,
        data,
        *,
        raises=True,
        minkw={},
        gpfactorykw={},
        jit=True,
        method='gradient',
        initial='priormean',
        verbosity=0,
        covariance='auto',
        fix=None,
        mlkw={},
        forward=False,
        additional_loss=None,
    ):
        """
    
        Maximum a posteriori fit.
    
        Maximizes the marginal likelihood of the data with a Gaussian process
        model that depends on hyperparameters, multiplied by a prior on the
        hyperparameters.
    
        Parameters
        ----------
        hyperprior : scalar, array or dictionary of scalars/arrays
            A collection of gvars representing the prior for the
            hyperparameters.
        gpfactory : callable
            A function with signature gpfactory(hyperparams) -> GP object. The
            argument ``hyperparams`` has the same structure of the
            `empbayes_fit` argument ``hyperprior``. gpfactory must be
            JAX-friendly, i.e., use `jax.numpy` and `jax.scipy` instead of plain
            `numpy`/`scipy` and avoid assignments to arrays.
        data : dict, tuple or callable
            Dictionary of data that is passed to `GP.marginal_likelihood` on
            the GP object returned by ``gpfactory``. If a tuple, it contains the
            first two arguments to `GP.marginal_likelihood`. If a callable, it
            is called with the same arguments of ``gpfactory`` and must return
            the argument(s) for `GP.marginal_likelihood`.
        raises : bool, optional
            If True (default), raise an error when the minimization fails.
            Otherwise, use the last point of the minimization as result.
        minkw : dict, optional
            Keyword arguments passed to `scipy.optimize.minimize`, overwrites
            values specified by `empbayes_fit`.
        gpfactorykw : dict, optional
            Keyword arguments passed to ``gpfactory``, and also to ``data`` if
            it is a callable. If ``jit``, ``gpfactorykw`` crosses a `jax.jit`
            boundary, so it must contain objects understandable by `jax`.
        jit : bool
            If True (default), use `jax.jit` to compile the minimization target.
        method : str
            Minimization strategy. Options:
        
            'nograd'
                Use a gradient-free method.
            'gradient' (default)
                Use a gradient-only method.
            'fisher'
                Use a Newton method with the Fisher information matrix plus
                the hyperprior precision matrix.
        initial : str, scalar, array, dictionary of scalars/arrays
            Starting point for the minimization, matching the format of
            ``hyperprior``, or one of the following options:
            
            'priormean' (default)
                Start from the hyperprior mean.
            'priorsample'
                Take a random sample from the hyperprior.
        verbosity : int
            An integer indicating how much information is printed on the
            terminal:
    
            0 (default)
                No logging.
            1
                Minimal report.
            2
                Detailed report.
            3
                Log each iteration.
            4
                More detailed iteration log.
            5
                Print the current parameter values at each iteration.
        covariance : str
            Method to estimate the posterior covariance matrix of the
            hyperparameters:
    
            'fisher'
                Use the Fisher information in the MAP, plus the prior precision,
                as precision matrix.
            'minhess'
                Use the hessian estimate of the minimizer as precision matrix.
            'none'
                Do not estimate the covariance matrix.
            'auto' (default)
                ``'minhess'`` if applicable, ``'none'`` otherwise.
        fix : scalar, array or dictionary of scalars/arrays
            A set of booleans, with the same format as ``hyperprior``,
            indicating which hyperparameters are kept fixed to their initial
            value. Scalars and arrays are broadcasted to the shape of
            ``hyperprior``. If a dictionary, missing keys are treated as False.
        mlkw : dict
            Additional arguments passed to `GP.marginal_likelihood`.
        forward : bool, default False
            Use forward instead of backward derivatives. Typically, forward is
            faster with a small number of parameters.
        additional_loss : callable, optional
            A function with signature ``additional_loss(hyperparams) -> float``
            which is added to the minus log marginal posterior of the
            hyperparameters.
    
        Attributes
        ----------
        p : scalar, array or dictionary of scalars/arrays
            A collection of gvars representing the hyperparameters that
            maximize their posterior. These gvars do not track correlations
            with the hyperprior or the data.
        prior : scalar, array or dictionary of scalars/arrays
            A copy of the hyperprior.
        initial : scalar, array or dictionary of scalars/arrays
            Starting point of the minimization, with the same format as ``p``.
        fix : scalar, array or dictionary of scalars/arrays
            A set of booleans, with the same format as ``p``, indicating which
            parameters were kept fixed to the values in ``initial``.
        pmean : scalar, array or dictionary of scalars/arrays
            Mean of ``p``.
        pcov : scalar, array or dictionary of scalars/arrays
            Covariance matrix of ``p``.
        minresult : scipy.optimize.OptimizeResult
            The result object returned by `scipy.optimize.minimize`.
        minargs : dict
            The arguments passed to `scipy.optimize.minimize`.
        gpfactory : callable
            The ``gpfactory`` argument.
        gpfactorykw : dict
            The ``gpfactorykw`` argument.
        data : dict, tuple or callable
            The ``data`` argument.

        Raises
        ------
        RuntimeError
            The minimization failed and ``raises`` is True.
    
        """

        Logger.__init__(self, verbosity)
        del verbosity
        self.log('**** call lsqfitgp.empbayes_fit ****')
    
        assert callable(gpfactory)
        
        # analyze the hyperprior
        hpinitial, hpunflat = self._parse_hyperprior(hyperprior, initial, fix)
        del hyperprior, initial, fix
        
        # analyze data
        data, cachedargs = self._parse_data(data)

        # define functions
        timer, functions = self._prepare_functions(
            gpfactory=gpfactory, gpfactorykw=gpfactorykw, data=data,
            cachedargs=cachedargs, hpunflat=hpunflat, mlkw=mlkw, jit=jit,
            forward=forward, additional_loss=additional_loss,
        )
        del gpfactory, gpfactorykw, data, cachedargs, mlkw, forward, additional_loss

        # prepare minimizer arguments
        minargs = self._prepare_minargs(method, functions, hpinitial)

        # set up callback to time and log iterations
        callback = self._Callback(self, functions, timer, hpunflat)
        minargs.update(callback=callback)

        # check invalid argument before running minimizer
        if not covariance in ('auto', 'fisher', 'minhess', 'none'):
            raise KeyError(covariance)
        
        # add user arguments and minimize
        minargs.update(minkw)
        self.log(f'minimizer method {minargs["method"]!r}', 2)
        total = time.perf_counter()
        result = optimize.minimize(**minargs)

        # check the minimization was successful
        self._check_success(result, raises)
        
        # compute posterior covariance of the hyperparameters
        cov = self._posterior_covariance(method, covariance, result, functions['fisher'])    
        
        # log total timings and function calls
        total = time.perf_counter() - total
        self._log_totals(total, timer, callback, jit, functions)
        
        # join posterior mean and covariance matrix
        uresult = gvar.gvar(result.x, cov)
        
        # set attributes
        self.p = gvar.gvar(hpunflat(uresult))
        self.pmean = gvar.mean(self.p)
        self.pcov = gvar.evalcov(self.p)
        self.minresult = result
        self.minargs = minargs

        # tabulate hyperparameter prior and posterior
        if self._verbosity >= 2:
            self.log(_gvarext.tabulate_together(
                self.prior, self.p,
                headers=['param', 'prior', 'posterior'],
            )) # TODO replace tabulate_toegether with something more flexible I
            # can use for the callback as well. Maybe import TextMatrix from
            # miscpy.
            # TODO print the transformed parameters
        
        self.log('**** exit lsqfitgp.empbayes_fit ****')

    class _CountCalls:
        """ wrap a callable to count calls """
        
        def __init__(self, func):
            self._func = func
            self._total = 0
            self._partial = 0
            functools.update_wrapper(self, func)
        
        def __call__(self, *args, **kw):
            self._total += 1
            self._partial += 1
            return self._func(*args, **kw)
        
        def partial(self):
            """ return the partial counter and reset it """
            result = self._partial
            self._partial = 0
            return result
        
        def total(self):
            """ return the total number of calls """
            return self._total

        @staticmethod
        def fmtcalls(method, functions):
            """
            format summary of number of calls
            method : str
            functions: dict[str, _CountCalls]
            """
            def counts():
                for name, func in functions.items():
                    if count := getattr(func, method)():
                        yield f'{name} {count}'
            return ', '.join(counts())

    class _Timer:
        """ object to time likelihood computations """

        def __init__(self):
            self.totals = {}
            self.partials = {}
            self._last_start = False

        def start(self, token):
            return token_map(self._start, token)

        def _start(self, token):
            self.stamp = time.perf_counter()
            self.counter = 0
            assert not self._last_start # forbid consecutive start() calls
            self._last_start = True
            return token

        def reset(self):
            self.partials = {}

        def partial(self, token):
            return token_map(self._partial, token)

        def _partial(self, token):
            now = time.perf_counter()
            delta = now - self.stamp
            self.partials[self.counter] = self.partials.get(self.counter, 0) + delta
            self.totals[self.counter] = self.totals.get(self.counter, 0) + delta
            self.stamp = now
            self.counter += 1
            self._last_start = False
            return token
    
    def _parse_hyperprior(self, hyperprior, initial, fix):
        
        # check fix against hyperprior and fill missing values
        hyperprior = self._copyasarrayorbufferdict(hyperprior)
        self._check_no_redundant_keys(hyperprior)
        fix = self._parse_fix(hyperprior, fix)
        flatfix = self._flatview(fix)

        # extract distribution of free hyperparameters
        flathp = self._flatview(hyperprior)
        freehp = flathp[~flatfix]
        mean = gvar.mean(freehp)
        cov = gvar.evalcov(freehp) # TODO use evalcov_blocks
        dec = _linalg.Chol(cov)
        assert dec.n == freehp.size
        self.log(f'{freehp.size}/{flathp.size} free hyperparameters', 2)
        
        # determine starting point for minimization
        initial = self._parse_initial(hyperprior, initial, dec)
        flatinitial = self._flatview(initial)
        x0 = dec.pinv_correlate(flatinitial[~flatfix] - mean)
        # TODO for initial = 'priormean', x0 is zero, skip decorrelate
        # for initial = 'priorsample', x0 is iid normal, but I have to sync
        # it with the user-exposed unflattened initial in _parse_initial
        
        # make function to correlate, add fixed values, and reshape to original
        # format
        fixed_indices, = jnp.nonzero(flatfix)
        unfixed_indices, = jnp.nonzero(~flatfix)
        fixed_values = jnp.asarray(flatinitial[flatfix])
        def unflat(x):
            assert x.ndim == 1
            if x.dtype == object:
                jac, indices = _gvarext.jacobian(x)
                xmean = mean + dec.correlate(gvar.mean(x))
                xjac = dec.correlate(jac)
                x = _gvarext.from_jacobian(xmean, xjac, indices)
                y = numpy.empty(flatfix.size, x.dtype)
                numpy.put(y, unfixed_indices, x)
                numpy.put(y, fixed_indices, fixed_values)
            else:
                x = mean + dec.correlate(x)
                y = jnp.empty(flatfix.size, x.dtype)
                y = y.at[unfixed_indices].set(x)
                y = y.at[fixed_indices].set(fixed_values)
            return self._unflatview(y, hyperprior)
        
        self.prior = hyperprior
        return x0, unflat
    
    @staticmethod
    def _check_no_redundant_keys(hyperprior):
        if not hasattr(hyperprior, 'keys'):
            return
        for k in hyperprior:
            m = hyperprior.extension_pattern.match(k)
            if m and m.group(1) in hyperprior.invfcn:
                altk = m.group(2)
                if altk in hyperprior:
                    raise ValueError(f'duplicate keys {altk!r} and {k!r} in hyperprior')

    def _parse_fix(self, hyperprior, fix):
        
        if fix is None:
            if hasattr(hyperprior, 'keys'):
                fix = gvar.BufferDict(hyperprior, buf=numpy.zeros(hyperprior.size, bool))
            else:
                fix = numpy.zeros(hyperprior.shape, bool)
        else:
            fix = self._copyasarrayorbufferdict(fix)
            if hasattr(fix, 'keys'):
                assert hasattr(hyperprior, 'keys'), 'fix is dictionary but hyperprior is array'
                assert all(hyperprior.has_dictkey(k) for k in fix), 'some keys in fix are missing in hyperprior'
                newfix = {}
                for k, v in hyperprior.items():
                    key = None
                    m = hyperprior.extension_pattern.match(k)
                    if m and m.group(1) in hyperprior.invfcn:
                        altk = m.group(2)
                        if altk in fix:
                            assert k not in fix, f'duplicate keys {k!r} and {altk!r} in fix'
                            key = altk
                    if key is None and k in fix:
                        key = k
                    if key is None:
                        elem = numpy.zeros(v.shape, bool)
                    else:
                        elem = numpy.broadcast_to(fix[key], v.shape)
                    newfix[k] = elem
                fix = gvar.BufferDict(newfix, dtype=bool)
            else:
                assert not hasattr(hyperprior, 'keys'), 'fix is array but hyperprior is dictionary'
                fix = numpy.broadcast_to(fix, hyperprior.shape).astype(bool)
        
        self.fix = fix
        return fix
    
    def _parse_initial(self, hyperprior, initial, dec):
        
        if not isinstance(initial, str):
            self.log('start from provided point', 2)
            initial = self._copyasarrayorbufferdict(initial)
            if hasattr(hyperprior, 'keys'):
                assert hasattr(initial, 'keys'), 'hyperprior is dictionary but initial is array'
                assert set(hyperprior.keys()) == set(initial.keys())
                assert all(hyperprior[k].shape == initial[k].shape for k in hyperprior)
            else:
                assert not hasattr(initial, 'keys'), 'hyperprior is array but initial is dictionary'
                assert hyperprior.shape == initial.shape
        
        elif initial == 'priormean':
            self.log('start from prior mean', 2)
            initial = gvar.mean(hyperprior)
        
        elif initial == 'priorsample':
            self.log('start from a random sample from the prior', 2)
            if dec.n < hyperprior.size:
                flathp = self._flatview(hyperprior)
                cov = gvar.evalcov(flathp) # TODO use evalcov_blocks
                fulldec = _linalg.Chol(cov)
            else:
                fulldec = dec
            iid = numpy.random.randn(fulldec.m)
            flatinitial = numpy.asarray(fulldec.correlate(iid))
            initial = self._unflatview(flatinitial, hyperprior)
        
        else:
            raise KeyError(initial)
        
        self.initial = initial
        return initial
    
    def _parse_data(self, data):
        
        self.data = data
        if isinstance(data, tuple) and len(data) == 1:
            data, = data

        if callable(data):
            self.log('data is callable', 2)
            cachedargs = None
        elif isinstance(data, tuple):
            self.log('data errors provided separately', 2)
            assert len(data) == 2
            cachedargs = data
        elif (gdata := self._copyasarrayorbufferdict(data)).dtype == object:
            self.log('data has errors as gvars', 2)
            data = gvar.gvar(gdata)
            # convert to gvar because non-gvars in the array would upset
            # gvar.mean and gvar.evalcov
            cachedargs = (gvar.mean(data), gvar.evalcov(data))
        else:
            self.log('data has no errors', 2)
            cachedargs = (data,)
        
        return data, cachedargs

    def _prepare_functions(self, *, gpfactory, gpfactorykw, data, cachedargs,
        hpunflat, mlkw, jit, forward, additional_loss):

        timer = self._Timer()
        firstcall = [None]
        
        def make_decomp(p, **kw):
            """ decomposition of the prior covariance and data """

            # start timer and convert hypers to user format
            p = timer.start(p)
            hp = hpunflat(p)
            
            # create GP object
            gp = gpfactory(hp, **kw)
            assert isinstance(gp, _GP.GP)
            
            # extract data
            if cachedargs:
                args = cachedargs
            else:
                args = data(hp, **kw)
                if not isinstance(args, tuple):
                    args = (args,)

            # decompose covariance matrix and flatten data
            decomp, r = gp._prior_decomp(*args, covtransf=timer.partial, **mlkw)
            r = r.astype(float) # int data upsets jax
        
            # log number of datapoints
            if firstcall:
                # it is convenient to do here because the data is flattened.
                # works under jit since the first call is tracing
                firstcall.pop()
                xdtype = gp._get_x_dtype()
                nd = '?' if xdtype is None else _array._nd(xdtype)
                self.log(f'{r.size} datapoints, {nd} covariates')

            # compute user loss
            if additional_loss is None:
                loss = 0.
            else:
                loss = additional_loss(hp)

            # split timer and return decomposition
            return timer.partial(decomp), r, loss
                # TODO what's the correct way of checkpointing r?

        # define wrapper to collect call stats, pass user args, compile
        def wrap(func):
            if jit:
                func = jax.jit(func)
            func = functools.partial(func, **gpfactorykw)
            return self._CountCalls(func)
        if jit:
            self.log('compile functions with jax jit', 2)        
        
        # log derivation method
        modename = 'forward' if forward else 'reverse'
        self.log(f'{modename}-mode autodiff (if used)', 2)

        # TODO time the derivatives separately => maybe I need a custom
        # derivative rule for timer token acknoledgement?

        def prior(p):
            # the marginal prior of the hyperparameters is a Normal with
            # identity covariance matrix because p is transformed to make it so
            return 1/2 * (len(p) * jnp.log(2 * jnp.pi) + p @ p)

        def grad_prior(p):
            return p

        def fisher_prior(p):
            return jnp.eye(len(p))

        @wrap
        def fun(p, **kw):
            """ minus log marginal posterior of the hyperparameters (not
            normalized) """
            decomp, r, loss = make_decomp(p, **kw)
            cond, _, _, _, _ = decomp.minus_log_normal_density(r, value=True)
            post = cond + prior(p) + loss
                # TODO what's the correct way of checkpointing prior and loss?
            return timer.partial(post)

        def make_gradfwd_fisher_args(p, **kw):
            def make_decomp_tee(p):
                decomp, r, loss = make_decomp(p, **kw)
                return (decomp.matrix(), r, loss), (decomp, r, loss)
            (dK, dr, grad_loss), (decomp, r, loss) = jax.jacfwd(make_decomp_tee, has_aux=True)(p)
            lkw = dict(dK=dK, dr=dr)
            return decomp, r, lkw, loss, grad_loss

        def make_gradrev_args(p, **kw):
            def make_decomp_loss(p):
                def make_decomp_r(p):
                    def make_decomp_K(p):
                        decomp, r, loss = make_decomp(p, **kw)
                        return decomp.matrix(), (decomp, r, loss)
                    _, dK_vjp, (decomp, r, loss) = jax.vjp(make_decomp_K, p, has_aux=True)
                    return r, (decomp, r, dK_vjp, loss)
                _, dr_vjp, (decomp, r, dK_vjp, loss) = jax.vjp(make_decomp_r, p, has_aux=True)
                return loss, (decomp, r, dK_vjp, dr_vjp, loss)
            grad_loss, (decomp, r, dK_vjp, dr_vjp, loss) = jax.grad(make_decomp_loss, has_aux=True)(p)
            unpack = lambda f: lambda x: f(x)[0]
            dK_vjp = unpack(dK_vjp)
            dr_vjp = unpack(dr_vjp)
            lkw = dict(dK_vjp=dK_vjp, dr_vjp=dr_vjp)
            return decomp, r, lkw, loss, grad_loss

        def make_jac_args(p, **kw):
            if forward:
                out = make_gradfwd_fisher_args(p, **kw)
                out[2].update(gradfwd=True) # out[2] is lkw
            else:
                out = make_gradrev_args(p, **kw)
                out[2].update(gradrev=True)
            return out

        @wrap
        def fun_and_jac(p, **kw):
            """ fun and its gradient """
            decomp, r, lkw, loss, grad_loss = make_jac_args(p, **kw)
            cond, gradrev, gradfwd, _, _ = decomp.minus_log_normal_density(r, value=True, **lkw)
            post = cond + prior(p) + loss
            grad_cond = gradfwd if forward else gradrev
            grad_post = grad_cond + grad_prior(p) + grad_loss
            return timer.partial((post, grad_post))
        
        @wrap
        def jac(p, **kw):
            """ gradient of fun """
            decomp, r, lkw, _, grad_loss = make_jac_args(p, **kw)
            _, gradrev, gradfwd, _, _ = decomp.minus_log_normal_density(r, **lkw)
            grad_cond = gradfwd if forward else gradrev
            grad_post = grad_cond + grad_prior(p) + grad_loss
            return timer.partial(grad_post)

        @wrap
        def fisher(p, **kw):
            """ fisher matrix """
            if additional_loss is not None:
                raise NotImplementedError(
                    'Fisher matrix not implemented with additional_loss. It '
                    'is possible but I did not prioritize it. If you need it, '
                    'open an issue on github.')
            decomp, r, lkw, _, _ = make_gradfwd_fisher_args(p, **kw)
            _, _, _, fisher_cond, _ = decomp.minus_log_normal_density(r, fisher=True, **lkw)
            fisher_post = fisher_cond + fisher_prior(p)
            return timer.partial(fisher_post)
        
        # set attributes
        self.gpfactory = gpfactory
        self.gpfactorykw = gpfactorykw

        return timer, {
            'fun': fun,
            'jac': jac,
            'fun&jac': fun_and_jac,
            'fisher': fisher,
        }

    def _prepare_minargs(self, method, functions, hpinitial):
        minargs = dict(fun=functions['fun&jac'], jac=True, x0=hpinitial)
        if self.SEPARATE_JAC:
            minargs.update(fun=functions['fun'], jac=functions['jac'])
        if method == 'nograd':
            minargs.update(fun=functions['fun'], jac=None, method='nelder-mead')
        elif method == 'gradient':
            minargs.update(method='bfgs')
        elif method == 'fisher':
            minargs.update(hess=functions['fisher'], method='dogleg')
            # dogleg requires positive definiteness, fisher is p.s.d.
            # trust-constr has more options, but it seems to be slower than
            # dogleg, so I keep dogleg as default
        else:
            raise KeyError(method)
        self.log(f'method {method!r}', 2)
        return minargs

        # TODO add method with fisher matvec instead of fisher matrix

    def _log_totals(self, total, timer, callback, jit, functions):
        times = {
            'gp&cov': timer.totals[0],
            'decomp': timer.totals[1],
            'likelihood': timer.totals[2],
            'jit': None, # set now and delete later to keep it before 'other'
            'other': total - sum(timer.totals.values()),
        }
        if jit:
            overhead = callback.estimate_firstcall_overhead()
            # TODO this estimation ignores the jit compilation of the function
            # used to compute the precision matrix, to be precise I should
            # manually split the jit into compilation + evaluation or hook into
            # it somehow. Maybe the jit object keeps a compilation wall time
            # stat?
        if jit and overhead is not None:
            times['jit'] = overhead
            times['other'] -= overhead
        else:
            del times['jit']
        self.log('', 4)
        calls = self._CountCalls.fmtcalls('total', functions)
        self.log(f'calls: {calls}')
        self.log(f'total time: {callback.fmttime(total)}')
        self.log(f'partials: {callback.fmttimes(times)}', 2)

    def _check_success(self, result, raises):
        if result.success:
            self.log(f'minimization succeeded: {result.message}')
        else:
            msg = f'minimization failed: {result.message}'
            if raises:
                raise RuntimeError(msg)
            elif self._verbosity == 0:
                warnings.warn(msg)
            else:
                self.log(msg)

    def _posterior_covariance(self, method, covariance, minimizer_result, fisher_func):
        
        if covariance == 'auto':
            if hasattr(minimizer_result, 'hess_inv') or hasattr(minimizer_result, 'hess'):
                covariance = 'minhess'
            else:
                covariance = 'none'

        if covariance == 'fisher':
            self.log('use fisher plus prior precision as precision', 2)
            if method == 'fisher':
                prec = minimizer_result.hess
            else:
                prec = fisher_func(minimizer_result.x)
            cov = _linalg.Chol(prec).ginv()

        elif covariance == 'minhess':
            if hasattr(minimizer_result, 'hess_inv'):
                hessinv = minimizer_result.hess_inv
                if isinstance(hessinv, optimize.LbfgsInvHessProduct):
                    self.log(f'convert LBFGS({hessinv.n_corrs}) hessian inverse to BFGS as covariance', 2)
                    cov = self._invhess_lbfgs_to_bfgs(hessinv)
                    # TODO this still gives a too wide cov when the minimization
                    # terminates due to bad linear search, is it because of
                    # dropped updates? This is currently keeping me from setting
                    # l-bfgs-b as default minimization method.
                elif isinstance(hessinv, numpy.ndarray):
                    self.log('use minimizer estimate of inverse hessian as covariance', 2)
                    cov = hessinv
            elif hasattr(minimizer_result, 'hess'):
                self.log('use minimizer hessian as precision', 2)
                cov = _linalg.Chol(minimizer_result.hess).ginv()
            else:
                raise RuntimeError('the minimizer did not return an estimate of the hessian')

        elif covariance == 'none':
            cov = numpy.full(minimizer_result.x.size, numpy.nan)

        else:
            raise KeyError(covariance)

        return cov

    @staticmethod
    def _invhess_lbfgs_to_bfgs(lbfgs):
        bfgs = optimize.BFGS()
        bfgs.initialize(lbfgs.shape[0], 'inv_hess')
        for i in range(lbfgs.n_corrs):
            bfgs.update(lbfgs.sk[i], lbfgs.yk[i])
        return bfgs.get_matrix()

    class _Callback:
        """ Iteration callback for scipy.optimize.minimize """
        
        def __init__(self, this, functions, timer, unflat):
            self.it = 0
            self.stamp = time.perf_counter()
            self.this = this
            self.functions = functions
            self.timer = timer
            self.unflat = unflat
            self.tail_overhead = 0
            self.tail_overhead_iter = 0

        def __call__(self, intermediate_result, arg2=None):
            
            if isinstance(intermediate_result, optimize.OptimizeResult):
                p = intermediate_result.x
            elif isinstance(intermediate_result, numpy.ndarray):
                p = intermediate_result
            else:
                raise TypeError(type(intermediate_result))

            self.it += 1
            now = time.perf_counter()
            duration = now - self.stamp
            
            worktime = sum(self.timer.partials.values())
            if worktime:
                overhead = duration - worktime
                assert overhead >= 0, (duration, worktime)
                if self.it == 1:
                    self.first_overhead = overhead
                else:
                    self.tail_overhead_iter += 1
                    self.tail_overhead += overhead

            # level 3 log
            calls = self.this._CountCalls.fmtcalls('partial', self.functions)
            times = self.fmttime(duration)
            self.this.log(f'iter {self.it}, time: {times}, calls: {calls}', {3})

            # level 4 log
            tot = self.fmttime(duration)
            if self.timer.partials:
                times = {
                    'gp&cov': self.timer.partials[0],
                    'dec': self.timer.partials[1],
                    'like': self.timer.partials[2],
                    'other': duration - sum(self.timer.partials.values()),
                }
                times = self.fmttimes(times)
            else:
                times = 'n/d'
            self.this.log(f'\niteration {self.it}', 4)
            with self.this.loglevel:
                self.this.log(f'total time: {tot}', 4)
                self.this.log(f'partial: {times}', 4)
                self.this.log(f'calls: {calls}', 4)

            # level 5 log
            nicep = self.unflat(p)
            nicep = self.this._copyasarrayorbufferdict(nicep)
            with self.this.loglevel:
                self.this.log(f'parameters = {nicep}', 5)
            # TODO write a method to format the parameters nicely. => use
            # gvar.tabulate? => nope, need actual gvars
            # TODO does this logging add significant overhead?

            self.stamp = now
            self.timer.reset()

        pattern = re.compile(
            r'((\d+) days, )?(\d{1,2}):(\d\d):(\d\d(\.\d{6})?)')

        @classmethod
        def fmttime(cls, seconds):
            if seconds < 0:
                prefix = '-'
                seconds = -seconds
            else:
                prefix = ''
            return prefix + cls._fmttime_positive(seconds)

        @classmethod
        def _fmttime_positive(cls, seconds):
            td = datetime.timedelta(seconds=seconds)
            m = cls.pattern.fullmatch(str(td))
            _, day, hour, minute, second, _ = m.groups()
            hour = int(hour)
            minute = int(minute)
            second = float(second)
            if day:
                return f'{day.lstrip("0")}d{hour:02d}h'
            elif hour:
                return f'{hour}h{minute:02d}m'
            elif minute:
                return f'{minute}m{second:02.0f}s'
            elif second >= 0.0995:
                return f'{second:#.2g}'.rstrip('.') + 's'
            elif second >= 0.0000995:
                return f'{second * 1e3:#.2g}'.rstrip('.') + 'ms'
            else:
                return f'{second * 1e6:.0f}μs'

        @classmethod
        def fmttimes(cls, times):
            """ times = dict label -> seconds """
            return ', '.join(f'{k} {cls.fmttime(v)}' for k, v in times.items())

        def estimate_firstcall_overhead(self):
            if self.tail_overhead_iter and hasattr(self, 'first_overhead'):
                typical_overhead = self.tail_overhead / self.tail_overhead_iter
                return self.first_overhead - typical_overhead

    @staticmethod
    def _copyasarrayorbufferdict(x):
        if hasattr(x, 'keys'):
            return gvar.BufferDict(x)
        else:
            return numpy.array(x)

    @staticmethod
    def _flatview(x):
        if hasattr(x, 'reshape'):
            return x.reshape(-1)
        elif hasattr(x, 'buf'):
            return x.buf
        else: # pragma: no cover
            raise NotImplementedError

    @staticmethod
    def _unflatview(x, original):
        if isinstance(original, numpy.ndarray):
            # TODO is this never applied to jax arrays?
            out = x.reshape(original.shape)
            # if not out.shape:
            #     try:
            #         out = out.item()
            #     except jax.errors.ConcretizationTypeError:
            #         pass
            return out
        elif isinstance(original, gvar.BufferDict):
            # normally I would do BufferDict(original, buf=x) but it does not
            # work with JAX tracers
            b = gvar.BufferDict(original)
            b._extension = {}
            b._buf = x
            # b.buf = x does not work because BufferDict checks that the
            # array is a numpy array
            # TODO maybe make a feature request to gvar to accept array_like
            # buf
            return b
        else: # pragma: no cover
            raise NotImplementedError


# TODO would it be meaningful to add correlation of the fit result with the data
# and hyperprior?

# TODO add the second order correction. It probably requires more than the
# gradient and inv_hess, but maybe by getting a little help from
# marginal_likelihood I can use the least-squares optimized second order
# correction on the residuals term and invent something for the logdet term.

# TODO it raises very often with "Desired error not necessarily achieved due to
# precision loss.". I tried doing a forward grad on the logdet but does not fix
# the problem. I still suspect it's the logdet, maybe the value itself and not
# the derivative, because as the matrix changes the regularization can change a
# lot the value of the logdet. How do I stabilize it? => scipy's l-bfgs-b seems
# to fail the linear search less often

# TODO compute the logGBF for the whole fit (see the gpbart code). In its doc,
# specify that 1) additional_loss may break the normalization if the user does
# not know what they are doing 2) the calculation of the log determinant term
# heavily depends on the regularization if the covariance matrix is singular;
# this won't happen if there are independent error terms in the model as usual.

# TODO empbayes_fit(autoeps=True) tries to double epsabs until the minimization
# succedes, with some maximum number of tries. autoeps=dict(maxrepeat=5,
# increasefactor=2, initial=1e-16, startfromzero=True) allows to configure the
# algorithm.

# TODO empbayes_fit(maxiter=100) sets the maximum number of minimization
# iterations. maxiter=dict(iter=100, calls=200, callsperiter=10) allows to
# configure it more finely. The calls limits are cumulative on all functions
# (need to make a class counter in _CountCalls), I can probably implement them
# by returning nan when the limit is surpassed, I hope the minimizer stops
# immediately on nan (test this). => Callback can raise StopIteration.

# TODO can I approximate the hessian with only function values and no gradient,
# i.e., when using nelder-mead? => See Hare (2022), although I would not know
# how to apply it properly to the optimization history. Somehow I need to keep
# only the "last" iterations.

# TODO is there a better algorithm than lbfgs for inaccurate functions? consider
# SC-BFGS (https://github.com/frankecurtis/SCBFGS). See Basak (2022). And NonOpt
# (https://github.com/frankecurtis/NonOpt).

# TODO can I estimate the error on the likelihood with the matrices? It requires
# the condition number. Basak (2022) gives wide bounds. I could try an upper
# bound and see how it compares to the true error, assuming that the matrix was
# as ill-conditioned as possible, i.e., use eps as the lowest eigenvalue, and
# gershgorin as the highest one.

# TODO look into jaxopt: it has improved a lot since the last time I saw it. In
# particular, it implements l-bfgs and has a "do not stop on failed line search"
# option. And it probably supports float32, although a skim of the docs suggests
# it does not work well. => See also optimistix.

# TODO reimplement the timing system with host_callback.id_tap. It should
# preserve the order because id_tap takes inputs and outputs. I must take care
# to make all callbacks happen at runtime instead of having some of them at
# compile time. I tried once but failed. Currently host_callback is
# experimental, maybe wait until it isn't. => I think it fails because it's
# asynchronous and there is only one device. Maybe host_callback.call would
# work? => I think they are developing something like my token machinery.

# TODO dictionary argument jitkw, arguments passed to jax.jit?

# TODO parameter float32: bool to use short float type. I think that scipy's
# optimize may break down with short floats with default options, I hope that
# changing termination tolerances does the trick.

# TODO make separate_jac a parameter

# TODO add options in _CountCalls to track inputs and/or outputs to some maximum
# buffer length, activate it if the method (after applying user options,
# lowercasing, and inferring minimize's default) is l-bfgs-b and the covariance
# is minhess or auto, to the order specified in the arguments to l-bfgs-b (after
# defaults inference if missing) (add tests in test_fit to check that the
# defaults stay as inferred), to be used if l-bfgs-b returns a crooked hessian.
# --- Alternative: if covariance = 'auto', it could be appropriate to use fisher
# per definition. --- Alternative: add option covariance = 'lbfgs(<order>)' that
# does this for any method, although this would require computing the gradients
# afterwards if the gradient was not used. These alternatives are not mutually
# exclusive.

# TODO make a helper function/class method that takes in data transf dependent
# on hypers and outputs additional loss (the log jacobian of the appropriate
# function with the appropriate sign)



# lsqfitgp/_gvarext/_tabulate.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import textwrap

import gvar
import numpy

def tabulate_together(*gs, headers=True, offset='', ndecimal=None, keys=None):
    """
    
    Format a table comparing side by side various collections of gvars.
    
    Parameters
    ----------
    *gs : sequence of arrays or dictionaries of gvars
        The variables to be tabulated. The structures of arrays and dictionaries
        must match.
    headers : bool or sequence of strings
        If True (default), add automatically an header. If False, don't add an
        header. If a sequence with length len(gs) + 1, it contains the column
        names for the keys/indices and for each set of variables.
    offset : str
        Prefix to each line, default empty.
    ndecimal : int, optional
        Number of decimal places. If not specified (default), keep two error
        digits.
    keys : sequence, optional
        If ``gs`` are dictionaries, a subset of keys to be extracted from each
        dictionary. Ignored if they are arrays.
    
    Examples
    --------
    >>> print(tabulate_together(gvar.gvar(dict(a=1)), gvar.gvar(dict(a=2))))
    key/index   value1   value2
    ---------------------------
            a    1 (0)    2 (0)
    
    See also
    --------
    gvar.tabulate
    
    """
    if not gs:
        return ''
    gs = [g if hasattr(g, 'keys') else numpy.asarray(g) for g in gs]
    assert all(hasattr(g, 'keys') for g in gs) or all(not hasattr(g, 'keys') for g in gs)
    if keys is not None and hasattr(gs[0], 'keys'):
        gs = [{k: g[k] for k in keys} for g in gs]
    g0 = gs[0]
    if hasattr(g0, 'keys'):
        assert all(set(g.keys()) == set(g0.keys()) for g in gs[1:])
        gs = [{k: g[k] for k in g0} for g in gs]
    else:
        assert all(g.shape == g0.shape for g in gs[1:])
        if g0.shape == ():
            gs = [{'--': g} for g in gs]
    tables = [
        _splittable(gvar.tabulate(g, headers=['@', ''], ndecimal=ndecimal))
        for g in gs
    ]
    columns = list(tables[0]) + [t[1] for t in tables[1:]]
    if not hasattr(headers, '__len__'):
        if headers:
            headers = ['key/index'] + [f'value{i+1}' for i in range(len(gs))]
        else:
            headers = None
    else:
        assert len(headers) == len(columns)
    if headers is not None:
        columns = (_head(col, head) for col, head in zip(columns, headers))
    return textwrap.indent(_join(columns), offset)

def _splittable(table):
    lines = table.split('\n')
    header = lines[0]
    col = header.find('@') + 1
    contentlines = lines[2:]
    col1 = '\n'.join(line[:col] for line in contentlines)
    col2 = '\n'.join(line[col:] for line in contentlines)
    return col1, col2

def _head(col, head):
    head = str(head)
    width = col.find('\n')
    if width < 0:
        width = len(col)
    hwidth = len(head)
    if hwidth > width:
        col = textwrap.indent(col, (hwidth - width) * ' ')
    else:
        head = (width - hwidth) * ' ' + head
    return head + '\n' + len(head) * '-' + '\n' + col

def _join(cols):
    split = (col.split('\n') for col in cols)
    return '\n'.join(''.join(lines) for lines in zip(*split))



# lsqfitgp/_Kernel/__init__.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from ._util import prod_recurse_dtype, sum_recurse_dtype, is_numerical_scalar
from ._crosskernel import CrossKernel, AffineSpan, PreservedBySwap
from . import _ops # keep first
from . import _alg # keep first
from ._kernel import Kernel
from ._stationary import CrossStationaryKernel, StationaryKernel
from ._isotropic import CrossIsotropicKernel, IsotropicKernel, Zero
from ._decorators import (crosskernel, kernel, crossstationarykernel,
    stationarykernel, crossisotropickernel, isotropickernel)



# lsqfitgp/_special/_expint.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools

from scipy import special
import jax
from jax import numpy as jnp
from jax.scipy import special as jspecial

from . import _gamma
from . import _taylor
from .. import _jaxext

@functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
def expn_imag(n, x):
    """
    Compute E_n(-ix), n integer >= 2, x real >= 0
    """
    
    # expn_imag_smallx loses accuracy due to cancellation between two terms
    # ~ x^n-2, while the result ~ x^-1, thus the relative error ~ x^-1/x^n-2 =
    # = x^-(n-1)
    #
    # error of expn_imag_smallx: eps z^n-1 E_1(z) / Gamma(n) ~
    #                            ~ eps z^n-2 / Gamma(n)
    #
    # error of expn_asymp: e^-z/z (n)_nt e^z/z^nt-1 E_n+nt(z) =
    #                      = (n)_nt / z^nt E_n+nt(z) ~
    #                      ~ (n)_nt / z^nt+1
    #
    # set the errors equal:
    #   eps z^n-2 / Gamma(n) = (n)_nt / z^nt+1  -->
    #   -->  z = (Gamma(n + nt) / eps)^1/(n+nt-1)
    
    # TODO improve accuracy at large n, it is probably sufficient to use
    # something like softmin(1/(n-1), 1/x) e^-ix, where the softmin scale
    # increases with n (how?)
    
    x = jnp.asarray(x)
    with jax.ensure_compile_time_eval():
        n = jnp.asarray(n)
        dt = _jaxext.float_type(n, x)
        if dt == jnp.float32:
            nt = jnp.array(10, 'i4') # TODO optimize to raise maximum n
        else:
            nt = 20 # TODO optimize to raise maximum n
        eps = jnp.finfo(dt).eps
        knee = (special.gamma(n + nt) / eps) ** (1 / (n + nt - 1))
    small = expn_imag_smallx(n, x)
    large = expn_asymp(n, -1j * x, nt)
    return jnp.where(x < knee, small, large)

@expn_imag.defjvp
def expn_imag_jvp(n, primals, tangents):
    
    # DLMF 8.19.13
    
    x, = primals
    xt, = tangents
    return expn_imag(n, x), xt * 1j * expn_imag(n - 1, x)

def expn_imag_smallx(n, x):
        
    # DLMF 8.19.7
    
    n, x = jnp.asarray(n), jnp.asarray(x)
    k = jnp.arange(n)
    fact = jnp.cumprod(k.at[0].set(1), dtype=_jaxext.float_type(n, x))
    n_1fact = fact[-1]
    ix = 1j * x
    E_1 = exp1_imag(x) # E_1(-ix)
    E_1 = jnp.where(x, E_1, 0) # Re E_1(-ix) ~ log(x) for x -> 0
    part1 = ix ** (n - 1) * E_1
    coefs = fact[:-1][(...,) + (None,) * ix.ndim]
    part2 = jnp.exp(ix) * jnp.polyval(coefs, ix)
    return (part1 + part2) / n_1fact
    
    # TODO to make this work with jit n, since the maximum n is something
    # like 30, I can always compute all the terms and set some of them to zero

def expn_asymp_coefgen(s, e, n):
    k = jnp.arange(s, e, dtype=n.dtype)
    return (-1) ** k * _gamma.poch(n, k)

def expn_asymp(n, z, nt):
    """
    Compute E_n(z) for large |z|, |arg z| < 3/2 π. ``nt`` is the number of terms
    used in the asymptotic series.
    """
    
    # DLMF 8.20.2
    
    invz = 1 / z
    return jnp.exp(-z) * invz * _taylor.taylor(expn_asymp_coefgen, (n,), 0, nt, invz)

_si_num = [
    1,
    -4.54393409816329991e-2, # x^2
    1.15457225751016682e-3, # x^4
    -1.41018536821330254e-5, # x^6
    9.43280809438713025e-8, # x^8
    -3.53201978997168357e-10, # x^10
    7.08240282274875911e-13, # x^12
    -6.05338212010422477e-16, # x^14
]

_si_denom = [
    1,
    1.01162145739225565e-2, # x^2
    4.99175116169755106e-5, # x^4
    1.55654986308745614e-7, # x^6
    3.28067571055789734e-10, # x^8
    4.5049097575386581e-13, # x^10
    3.21107051193712168e-16, # x^12
]

_ci_num = [
    -0.25,
    7.51851524438898291e-3, # x^2
    -1.27528342240267686e-4, # x^4
    1.05297363846239184e-6, # x^6
    -4.68889508144848019e-9, # x^8
    1.06480802891189243e-11, # x^10
    -9.93728488857585407e-15, # x^12
]

_ci_denom = [
    1,
    1.1592605689110735e-2, # x^2
    6.72126800814254432e-5, # x^4
    2.55533277086129636e-7, # x^6
    6.97071295760958946e-10, # x^8
    1.38536352772778619e-12, # x^10
    1.89106054713059759e-15, # x^12
    1.39759616731376855e-18, # x^14
]

_f_num = [
    1,
    7.44437068161936700618e2, # x^-2
    1.96396372895146869801e5, # x^-4
    2.37750310125431834034e7, # x^-6
    1.43073403821274636888e9, # x^-8
    4.33736238870432522765e10, # x^-10
    6.40533830574022022911e11, # x^-12
    4.20968180571076940208e12, # x^-14
    1.00795182980368574617e13, # x^-16
    4.94816688199951963482e12, # x^-18
    -4.94701168645415959931e11, # x^-20
]

_f_denom = [
    1,
    7.46437068161927678031e2, # x^-2
    1.97865247031583951450e5, # x^-4
    2.41535670165126845144e7, # x^-6
    1.47478952192985464958e9, # x^-8
    4.58595115847765779830e10, # x^-10
    7.08501308149515401563e11, # x^-12
    5.06084464593475076774e12, # x^-14
    1.43468549171581016479e13, # x^-16
    1.11535493509914254097e13, # x^-18
]

_g_num = [
    1,
    8.1359520115168615e2, # x^-2
    2.35239181626478200e5, # x^-4
    3.12557570795778731e7, # x^-6
    2.06297595146763354e9, # x^-8
    6.83052205423625007e10, # x^-10
    1.09049528450362786e12, # x^-12
    7.57664583257834349e12, # x^-14
    1.81004487464664575e13, # x^-16
    6.43291613143049485e12, # x^-18
    -1.36517137670871689e12, # x^-20
]

_g_denom = [
    1,
    8.19595201151451564e2, # x^-2
    2.40036752835578777e5, # x^-4
    3.26026661647090822e7, # x^-6
    2.23355543278099360e9, # x^-8
    7.87465017341829930e10, # x^-10
    1.39866710696414565e12, # x^-12
    1.17164723371736605e13, # x^-14
    4.01839087307656620e13, # x^-16
    3.99653257887490811e13, # x^-18
]

def _si_smallx(x):
    """ Compute Si(x) = int_0^x dt sin t / t, for x < 4"""
    x2 = jnp.square(x)
    dtype = _jaxext.float_type(x)
    num = jnp.polyval(jnp.array(_si_num[::-1], dtype), x2)
    denom = jnp.polyval(jnp.array(_si_denom[::-1], dtype), x2)
    return x * num / denom

def _minus_cin_smallx(x):
    """ Compute -Cin(x) = int_0^x dt (cos t - 1) / t, for x < 4 """
    x2 = jnp.square(x)
    dtype = _jaxext.float_type(x)
    num = jnp.polyval(jnp.array(_ci_num[::-1], dtype), x2)
    denom = jnp.polyval(jnp.array(_ci_denom[::-1], dtype), x2)
    return x2 * num / denom

def _ci_smallx(x):
    """ Compute Ci(x) = -int_x^oo dt cos t / t, for x < 4 """
    gamma = 0.57721566490153286060
    return gamma + jnp.log(x) + _minus_cin_smallx(x)

def _f_largex(x):
    """ Compute f(x) = int_0^oo dt sin t / (x + t), for x > 4 """
    x2 = 1 / jnp.square(x)
    dtype = _jaxext.float_type(x)
    num = jnp.polyval(jnp.array(_f_num[::-1], dtype), x2)
    denom = jnp.polyval(jnp.array(_f_denom[::-1], dtype), x2)
    return num / denom / x

def _g_largex(x):
    """ Compute g(x) = int_0^oo dt cos t / (x + t), for x > 4 """
    x2 = 1 / jnp.square(x)
    dtype = _jaxext.float_type(x)
    num = jnp.polyval(jnp.array(_g_num[::-1], dtype), x2)
    denom = jnp.polyval(jnp.array(_g_denom[::-1], dtype), x2)
    return x2 * num / denom

def _exp1_imag_smallx(x):
    """ Compute E_1(-ix), for x < 4 """
    return -_ci_smallx(x) + 1j * (jnp.pi / 2 - _si_smallx(x))

def _exp1_imag_largex(x):
    """ Compute E_1(-ix), for x > 4 """
    s = jnp.sin(x)
    c = jnp.cos(x)
    f = _f_largex(x)
    g = _g_largex(x)
    real = -f * s + g * c
    imag = f * c + g * s
    return real + 1j * imag  # e^ix (g + if)

@jax.jit
def exp1_imag(x):
    """
    Compute E_1(-ix) = int_1^oo dt e^ixt / t, for x > 0
    Reference: Rowe et al. (2015, app. B)
    """
    return jnp.where(x < 4, _exp1_imag_smallx(x), _exp1_imag_largex(x))
    
    # TODO This is 40x faster than special.exp1(-1j * x) and 2x than
    # special.sici(x), and since the jit has to run (I'm guessing) through both
    # branches of jnp.where, a C/Cython implementation would be 4x faster. Maybe
    # PR it to scipy for sici, after checking the accuracy against mpmath and
    # the actual C performance.

    # Do Padé approximants work for complex functions?

@jax.custom_jvp
def ci(x):
    return -exp1_imag(x).real

@ci.defjvp
def _ci_jvp(primals, tangents):
    x, = primals
    xt, = tangents
    return ci(x), xt * jnp.cos(x) / x



# lsqfitgp/copula/_makedict.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" defines makedict """

import gvar

from . import _base

def makedict(variables, prefix='__copula_'):
    """

    Expand distributions in a dictionary.

    Parameters
    ----------
    variables : dict
        A dictionary representing a collection of probability distribution. If a
        value is an instance of `DistrBase`, the key is converted to mark a
        transformation and the value is replaced with new primary gvars.
    prefix : str
        A prefix to make the transformation names unique.

    Returns
    -------
    out : BufferDict
        The transformed dictionary. Recognizes the same keys as `variables`,
        but squashes the values through the transformation that sends a Normal
        to the desired distribution.

    Raises
    ------
    ValueError :
        If any `DistrBase` object appears under different keys.

    Examples
    --------

    Put a `Distr` into a `gvar.BufferDict`:

    >>> bd = lgp.copula.makedict({'x': lgp.copula.beta(1, 1)})
    >>> bd
    BufferDict({'__copula_beta{1, 1}(x)': 0.0(1.0)})
    >>> bd['x']
    0.50(40)
    >>> bd['__copula_beta{1, 1}(x)']
    0.0(1.0)

    You can also put an entire `Copula`:

    >>> bd = lgp.copula.makedict({
    ...     'x': lgp.copula.Copula({
    ...         'y': lgp.copula.gamma(2, 1 / 2),
    ...         'z': lgp.copula.invgamma(2, 2),
    ...     }),
    ... })
    >>> bd
    BufferDict({"__copula_{'y': gamma{2, 0.5}, 'z': invgamma{2, 2}}(x)": array([0.0(1.0), 0.0(1.0)], dtype=object)})
    >>> bd['x']
    {'y': 3.4(2.5), 'z': 1.19(90)}

    Other entries and transformations are left alone:

    >>> bd = lgp.copula.makedict({
    ...     'x': gvar.gvar(3, 0.2),
    ...     'log(y)': gvar.gvar(0, 1),
    ...     'z': lgp.copula.dirichlet(1.5, [1, 2, 3]),
    ... })
    >>> bd
    BufferDict({'x': 3.00(20), 'log(y)': 0.0(1.0), '__copula_dirichlet{1.5, [1, 2, 3], shape=3}(z)': array([0.0(1.0), 0.0(1.0), 0.0(1.0)], dtype=object)})
    >>> bd['z']
    array([0.06(20), 0.31(49), 0.63(51)], dtype=object)
    >>> bd['y']
    1.0(1.0)

    Since shared `DistrBase` objects represent statistical dependency, it is
    forbidden to have the same object appear under different keys, as that
    would make it impossible to take the dependencies into account:

    >>> x = lgp.copula.beta(1, 1)
    >>> y = lgp.copula.beta(1, x)
    >>> lgp.copula.makedict({'x': x, 'y': y})
    ValueError: cross-key occurrences of object(s):
    beta with id 10952248976: <x>, <y.1>

    """

    # collect all objects and their representations in DistrBase instances
    caches = {}
    for k, v in variables.items():
        if isinstance(v, _base.DistrBase):
            cache = {}
            v.__repr__(k, cache)
            caches[k] = cache

    # put everything into a single multiple valued dict
    allobjects = {}
    for cache in caches.values():
        for obj, descr in cache.items():
            allobjects.setdefault(obj, []).append(descr)

    # find objects that appear multiple times
    multiple = ''
    for obj, descrs in allobjects.items():
        if len(descrs) > 1:
            multiple += f'{obj.__class__.__name__} with id {id(obj)}: {", ".join(descrs)}\n'

    # raise an error if there are
    if multiple:
        raise ValueError(f'cross-key occurrences of object(s):\n{multiple}')

    out = {}
    for k, v in variables.items():
        if isinstance(v, _base.DistrBase):
            name = str(v._staticdescr).replace('(', '{').replace(')', '}')
            assert '(' not in prefix and ')' not in prefix
                # gvar does not currently check presence of parentheses, see
                # https://github.com/gplepage/gvar/issues/39
            name = prefix + name
            v.add_distribution(name)
            v = v.gvars()
            k = f'{name}({k})'
        assert k not in out
        out[k] = v
    return gvar.BufferDict(out)





# lsqfitgp/_gvarext/__init__.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import contextlib

import gvar

from ._jacobian import jacobian, from_jacobian
from ._tabulate import tabulate_together
from ._ufunc import gvar_gufunc
from ._format import uformat, fmtspec_kwargs, gvar_format

@contextlib.contextmanager
def switchgvar():
    """
    Context manager to keep new gvars in a separate pool.

    Creating new primary gvars fills up memory permanently. This context manager
    keeps the gvars created within its context in a separate pool that is freed
    when all such gvars are deleted. They can not be mixed in operations with
    other gvars created outside of the context.
    
    Returns
    -------
    gvar : gvar.GVarFactory
        The new gvar-creating function that uses a new pool. The change is also
        reflected in the global `gvar.gvar`.

    See also
    --------
    gvar.switch_gvar, gvar.restore_gvar

    Examples
    --------

    >>> x = gvar.gvar(0, 1)
    >>> with lgp.switchgvar():
    >>>     y = gvar.gvar(0, 1)
    >>>     z = gvar.gvar(0, 1)
    >>> w = gvar.gvar(0, 1)
    >>> q = y + z  # allowed, y and z created in the same pool
    >>> p = x + w  # allowed, x and w created in the same pool
    >>> h = x + y  # x and y created in different pools: this will silently
    ...            # fail and possibly crash python immediately or later on

    """
    try:
        yield gvar.switch_gvar()
    finally:
        gvar.restore_gvar()



# lsqfitgp/_linalg/_decomp.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

Copy-pasted from the notes:

2023-02-14
==========

My current decomposition system is a mess. I can't take reverse gradients.
I can't straightforwardly implement optimized algorithms that compute together
likelihood, gradient, and fisher. Jax patterns break down unpredictably. I
have to redesign it from scratch.

Guidelines and requirements:

  - Sufficient modularity to implement composite decompositions (Woodbury,
    Block)

  - Does not mess up jax in any way

  - Caches decompositions

  - Favors optimizing together the likelihood and its derivatives

Operations (in the following I indicate with lowercase inputs which are
typically vectors or tall matrices, and uppercase inputs which are typically
large matrices, since optimization requires taking it into account):

    pinv_bilinear(A, r) => A'K⁺r (for the posterior mean)
    pinv_bilinear_robj(A, r) same but r can be gvars
    ginv_quad(A) => A'K⁻A (for the posterior covariance)
        I want the pseudoinverse for the mean because the data may not be
        in the span and I want to project it orthogonally, while for the
        covariance I expect A and K to come from a pos def matrix so they are
        coherent
    ginv_diagquad(A) => diag(A'K⁻A) (for the posterior variance)
    minus_log_normal_density(
        r: 1d array,      # the residuals (data - prior mean)
        dr_vjp: callable, # x -> x_i ∂r_i/∂p_j,   gradrev and fishvec
        dK_vjp: callable, # x -> x_ij ∂K_ij/∂p_k, gradrev and fishvec
        dr_jvp: callable, # x -> ∂r_i/∂r_j x_j,  fishvec
        dK_jvp: callable, # x -> ∂K_ij/∂p_k x_k, fishvec
        dr: 2d array,     # ∂r_i/∂p_j,  gradfwd and fisher
        dK: 3d array,     # ∂K_ij/∂p_k, gradfwd and fisher
        vec: 1d array,    # input vector of fishvec, same size as params
        value: bool,
        gradrev: bool,
        gradfwd: bool,
        fisher: bool,
        fishvec: bool,
    )
        This computes on request
            value: 1/2 tr(KK⁺) log 2π
                 + 1/2 tr(I-KK⁺) log 2π
                 + 1/2 log pdet K
                 + 1/2 tr(I-KK⁺) log ε
                 + 1/2 r'(K⁺+(I-KK⁺)/ε)r
            gradrev,
            gradfwd: 1/2 tr(K⁺dK)
                    + r'(K⁺+(I-KK⁺)/ε) dr
                    - 1/2 r'(K⁺+2(I-KK⁺)/ε)dKK⁺r
            fisher: 1/2 tr(K⁺dK(K⁺+2(I-KK⁺)/ε)d'K)
                  - 2 tr(K⁺dK(I-KK⁺)d'KK⁺)
                  + dr'(K⁺+(I-KK⁺)/ε)d'r
            fishvec: fisher matrix times vec
        There should be options for omitting the pieces with ε. I also need a
        way to make densities with different values of ε comparable with each
        other (may not be possible, if it is, it probably requires a history of
        ranks and ε). gradfwd/rev form K⁺ explicitly to compute tr(K⁺dK) for
        efficiency.
    correlate(x)
        Zx where K = ZZ'.
    back_correlate(X):
        Z'X, this is used by Sandwich and Woodbury.

Since I also want to compute the Student density, I could split
minus_log_normal_density's return value into logdet and quad. The gradient
splits nicely between the two terms, but I have to redo the calculation of the
Fisher matrix for the Student distribution. Alternatively, I could use the the
Normal Fisher. => See Lange et al. (1989, app. B). => I think I can split the
gradient and Fisher matrix too.

2023-03-07
==========

To compute a Fisher-vector product when there are many parameters, do

    tr(K+ dK K+ dK) v =
    = K_vjp(K+ K_jvp(v) K+)

"""

# TODO to automatize this further, I could take in a function that generates K
# (or its pieces) and the arguments to the function. But how would this play
# together with passing decomposition objects as pieces?

# TODO split this file by class

# TODO Consider using lineax for implementing non-materialized decomps

import abc
import functools

import numpy
import jax
from jax import numpy as jnp
from jax.scipy import linalg as jlinalg
from jax import lax

from .. import _jaxext
from . import _pytree

class Decomposition(_pytree.AutoPyTree, abc.ABC):
    """
    Abstract base class for decompositions of positive semidefinite matrices.
    """

    @abc.abstractmethod
    def __init__(self, *args, **kw):
        """ Decompose the input matrix """
        pass

    @abc.abstractmethod
    def matrix(self):
        """ The input matrix """
        pass

    @abc.abstractmethod
    def ginv_linear(self, X):
        """ Compute K⁻X """
        pass

    @abc.abstractmethod
    def pinv_bilinear(self, A, r):
        """Compute A'K⁺r."""
        pass

    @abc.abstractmethod
    def pinv_bilinear_robj(self, A, r):
        """Compute A'K⁺r, where r can be an array of objects."""
        pass

    @abc.abstractmethod
    def ginv_quad(self, A):
        """Compute A'K⁻A."""
        pass

    @abc.abstractmethod
    def ginv_diagquad(self, A):
        """Compute diag(A'K⁻A)."""
        pass

    @abc.abstractmethod
    def correlate(self, x):
        """ Compute Zx where K = ZZ' """
        pass

    @abc.abstractmethod
    def back_correlate(self, X):
        """ Compute Z'X """
        pass

    @abc.abstractmethod
    def pinv_correlate(self, x):
        """ Compute Z⁺x """
        pass

    @abc.abstractmethod
    def minus_log_normal_density(self,
        r,
        *,
        dr_vjp=None,
        dK_vjp=None,
        dr_jvp=None,
        dK_jvp=None,
        dr=None,
        dK=None,
        value=False,
        gradrev=False,
        gradfwd=False,
        fisher=False,
        fishvec=False,
        ):
        """
        Compute minus log a Normal density and its derivatives, with covariance
        matrix K.

        If an input derivative is not specified, it is assumed to be zero.

        Parameters
        ----------
        r: 1d array
            The residuals (value - mean)
        dr_vjp: callable
            x -> x_i ∂r_i/∂p_j, for gradrev and fishvec
        dK_vjp: callable
            x -> x_ij ∂K_ij/∂p_k, for gradrev and fishvec
        dr_jvp_vec: 1d array
            ∂r_i/∂r_j vec_j, for fishvec
        dK_jvp_vec: 2d array
            ∂K_ij/∂p_k vec_k, for fishvec
        dr: 2d array
            ∂r_i/∂p_j for gradfwd and fisher
        dK: 3d array
            ∂K_ij/∂p_k, for gradfwd and fisher
        value: bool
        gradrev: bool
        gradfwd: bool
        fisher: bool
        fishvec: bool
            These parameters indicate which of the return values to compute.
            Default all False.

        Returns
        -------
        value: 1/2 tr(KK⁺) log 2π
             + 1/2 tr(I-KK⁺) log 2π
             + 1/2 log pdet K
             + 1/2 tr(I-KK⁺) log ε
             + 1/2 r'(K⁺+(I-KK⁺)/ε)r
        gradrev,
        gradfwd: 1/2 tr(K⁺dK)
                + r'(K⁺+(I-KK⁺)/ε) dr
                - 1/2 r'(K⁺+2(I-KK⁺)/ε)dKK⁺r
        fisher: 1/2 tr(K⁺dK(K⁺+2(I-KK⁺)/ε)d'K)
              - 2 tr(K⁺dK(I-KK⁺)d'KK⁺)
              + dr'(K⁺+(I-KK⁺)/ε)d'r
        fishvec: fisher matrix @ vec
        """
        pass

    def _parseeps(self, K, epsrel, epsabs, maxeigv=None):
        """ Determine eps from input arguments """
        machine_eps = jnp.finfo(_jaxext.float_type(K)).eps
        if epsrel == 'auto':
            epsrel = len(K) * machine_eps
        if epsabs == 'auto':
            epsabs = machine_eps
        if maxeigv is None:
            maxeigv = eigval_bound(K)
        self._eps = epsrel * maxeigv + epsabs
        return self._eps

    @property
    def eps(self):
        """
        The threshold below which eigenvalues are too small to be determined.
        """
        return self._eps

    @property
    @abc.abstractmethod
    def n(self):
        """ Number of rows/columns of the matrix """
        pass

    @property
    @abc.abstractmethod
    def m(self):
        """ Number of columns of Z """
        pass

    def ginv(self):
        """ Compute K⁻ """
        return self.ginv_quad(jnp.eye(self.n))

def solve_triangular_python(a, b, *, lower=False):
    """
    Pure python implementation of scipy.linalg.solve_triangular for when
    a or b are object arrays.
    """
    # TODO maybe commit this to gvar.linalg
    a = numpy.asarray(a)
    x = numpy.copy(b)

    vec = x.ndim < 2
    if vec:
        x = x[:, None]

    n = a.shape[-1]
    assert x.shape[-2] == n

    if not lower:
        a = a[..., ::-1, ::-1]
        x = x[..., ::-1, :]

    x[..., 0, :] /= a[..., 0, 0, None]
    for i in range(1, n):
        x[..., i:, :] -= x[..., None, i - 1, :] * a[..., i:, i - 1, None]
        x[..., i, :] /= a[..., i, i, None]

    if not lower:
        x = x[..., ::-1, :]

    if vec:
        x = numpy.squeeze(x, -1)
    return x

def solve_triangular_batched(a, b, *, lower=False):
    """ Version of jax.scipy.linalg.solve_triangular that batches matmul-like """
    a = jnp.asarray(a)
    b = jnp.asarray(b)
    vec = b.ndim < 2
    if vec:
        b = b[:, None]

    batch_shape = jnp.broadcast_shapes(a.shape[:-2], b.shape[:-2])
    a_shape = batch_shape + a.shape[-2:]
    b_shape = batch_shape + b.shape[-2:]
    result = lax.linalg.triangular_solve(
        jnp.broadcast_to(a, a_shape), jnp.broadcast_to(b, b_shape),
        left_side=True, lower=lower,
    )
    assert result.shape == b_shape

    if vec:
        result = result.squeeze(-1)
    return result

def solve_batched(a, b, **kw):
    """ Version of jax.scipy.linalg.solve that batches matmul-like """
    a = jnp.asarray(a)
    b = jnp.asarray(b)
    vec = b.ndim < 2
    if vec:
        b = b[:, None]

    @functools.partial(jnp.vectorize, signature='(i,j),(j,k)->(i,k)')
    def solve_batched(a, b):
        return jlinalg.solve(a, b, **kw)
    result = solve_batched(a, b)

    if vec:
        result = result.squeeze(-1)
    return result

def eigval_bound(K):
    """
    Upper bound on the largest magnitude eigenvalue of the matrix, from
    Gershgorin's theorem.
    """
    return jnp.max(jnp.sum(jnp.abs(K), axis=1))

def diag_scale_pow2(K):
    """
    Compute a vector s of powers of 2 such that diag(K / outer(s, s)) ~ 1.
    """
    d = jnp.diag(K)
    return jnp.where(d, jnp.exp2(jnp.rint(0.5 * jnp.log2(d))), 1)

def transpose(x):
    """ swap the last two axes of array x, corresponds to matrix tranposition
    with the broadcasting convention of matmul """
    if x.ndim < 2:
        return x
    elif isinstance(x, jnp.ndarray):
        return jnp.swapaxes(x, -2, -1)
    else:
        # need to support numpy because this function is used with gvars
        return numpy.swapaxes(x, -2, -1)

class Chol(Decomposition):
    """Cholesky decomposition. The matrix is regularized adding a small multiple
    of the identity."""

    def __init__(self, K, *, epsrel='auto', epsabs=0):
        # K <- K + Iε
        # K = LL'
        self._K = K
        s = diag_scale_pow2(K)
        K = K / s / s[:, None]
        eps = self._parseeps(K, epsrel, epsabs)
        K = K.at[jnp.diag_indices_from(K)].add(eps)
        L = jlinalg.cholesky(K, lower=True)
        with _jaxext.skipifabstract():
            if not jnp.all(jnp.isfinite(L)):
                # TODO check that jax fills with nan after failed row, detect
                # and report minor index like scipy
                raise numpy.linalg.LinAlgError('cholesky decomposition not finite, probably matrix not pos def numerically')
        self._L = L * s[:, None]
        self._eps = eps * jnp.min(s * s)

    def matrix(self):
        return self._K

    def ginv_linear(self, X):
        # = K⁻¹X
        # K⁻¹ = L'⁻¹L⁻¹
        # K⁻¹X = L'⁻¹(L⁻¹X)
        invLX = jlinalg.solve_triangular(self._L, X, lower=True)
        return jlinalg.solve_triangular(self._L.T, invLX, lower=False)

    def pinv_bilinear(self, A, r):
        # = A'K⁻¹r = A'L'⁻¹L⁻¹r = (L⁻¹A)'(L⁻¹r)
        invLr = jlinalg.solve_triangular(self._L, r, lower=True)
        invLA = jlinalg.solve_triangular(self._L, A, lower=True)
        return invLA.T @ invLr

    def pinv_bilinear_robj(self, A, r):
        # = A'K⁻¹r
        invLr = solve_triangular_python(self._L, r, lower=True)
        invLA = jlinalg.solve_triangular(self._L, A, lower=True)
        return numpy.asarray(invLA).T @ invLr

    def ginv_quad(self, A):
        # = A'K⁻¹A = A'K⁻¹A = A'L'⁻¹L⁻¹A = (L⁻¹A)'(L⁻¹A)
        invLA = jlinalg.solve_triangular(self._L, A, lower=True)
        return invLA.T @ invLA

    def ginv_diagquad(self, A):
        # = diag(A'K⁻¹A)
        # X = L⁻¹A
        # diag(A'K⁻¹A)_i = diag(X'X)_i = ∑_j X'_ij X_ji = ∑_j X_ji X_ji
        invLA = jlinalg.solve_triangular(self._L, A, lower=True)
        return jnp.einsum('ji,ji->i', invLA, invLA)

    def correlate(self, x):
        # = Lx
        return self._L @ x

    def back_correlate(self, X):
        # = L'X
        return self._L.T @ X

    def pinv_correlate(self, x):
        # = L⁻¹x
        return jlinalg.solve_triangular(self._L, x, lower=True)

    def minus_log_normal_density(self,
        r,               # 1d array, the residuals (data - prior mean)
        *,
        dr_vjp=None,     # callable, x -> x_i ∂r_i/∂p_j,   gradrev and fishvec
        dK_vjp=None,     # callable, x -> x_ij ∂K_ij/∂p_k, gradrev and fishvec
        dr_jvp_vec=None, # 1d array, ∂r_i/∂r_j v_j,  fishvec
        dK_jvp_vec=None, # 2d array, ∂K_ij/∂p_k v_k, fishvec
        dr=None,         # 2d array, ∂r_i/∂p_j,  gradfwd and fisher
        dK=None,         # 3d array, ∂K_ij/∂p_k, gradfwd and fisher
        value=False,
        gradrev=False,
        gradfwd=False,
        fisher=False,
        fishvec=False,
    ):

        L = self._L

        out = {}

        # compute shared factors
        grad = (
            (gradrev and (dK_vjp is not None or dr_vjp is not None))
            or (gradfwd and (dK is not None or dr is not None))
        )
        if value or grad:
            invLr = jlinalg.solve_triangular(L, r, lower=True)
        if grad:
            invKr = jlinalg.solve_triangular(L.T, invLr, lower=False)
        if (gradrev and dK_vjp is not None) or (gradfwd and dK is not None):
            invL = jlinalg.solve_triangular(L, jnp.eye(len(L)), lower=True)
            invK = invL.T @ invL

        if value:
            # = 1/2 n log 2π
            #   + 1/2 log det K
            #   + 1/2 r'K⁻¹r
            # K = LL'
            # K⁻¹ = L'⁻¹L⁻¹
            # det K = (det L)² =
            #       = (∏_i L_ii)²
            # r'K⁻¹r = r'L'⁻¹L⁻¹r =
            #        = (L⁻¹r)'(L⁻¹r)
            out['value'] = 1/2 * (
                len(L) * jnp.log(2 * jnp.pi) +
                2 * jnp.sum(jnp.log(jnp.diag(L))) +
                invLr @ invLr
            )
        else:
            out['value'] = None

        if gradrev:
            # = 1/2 tr(K⁻¹dK)
            #   + r'K⁻¹dr
            #   - 1/2 r'K⁻¹dKK⁻¹r
            # tr(K⁻¹dK) = K⁻¹_ij dK_ji =
            #           = K⁻¹_ij dK_ij =
            #           = dK_vjp(K⁻¹)
            # r'K⁻¹dr = r_i K⁻¹_ij dr_j =
            #         = (K⁻¹r)_j dr_j =
            #         = dr_vjp(K⁻¹r)
            # r'K⁻¹dKK⁻¹r = r_i K⁻¹_ij dK_jl K⁻¹_lm r_m =
            #             = (K⁻¹r)_j dK_jl (K⁻¹r)_l =
            #             = dK_vjp((K⁻¹r) ⊗ (K⁻¹r))
            out['gradrev'] = 0
            if dK_vjp is not None:
                tr_invK_dK = dK_vjp(invK)
                r_invK_dK_invK_r = dK_vjp(jnp.outer(invKr, invKr))
                out['gradrev'] += 1/2 * (tr_invK_dK - r_invK_dK_invK_r)
            if dr_vjp is not None:
                r_invK_dr = dr_vjp(invKr)
                out['gradrev'] += r_invK_dr
        else:
            out['gradrev'] = None

        if gradfwd:
            # = 1/2 tr(K⁻¹dK)
            #   + r'K⁻¹dr
            #   - 1/2 r'K⁻¹dKK⁻¹r
            # tr(K⁻¹dK)_k = K⁻¹_ij dK_ijk
            # r'K⁻¹dr = (K⁻¹r)'dr
            # (r'K⁻¹dKK⁻¹r)_k = r_i K⁻¹_ij dK_jlk K⁻¹_lm r_m =
            #                 = (K⁻¹r)_j dK_jlk (K⁻¹r)_l
            out['gradfwd'] = 0
            if dK is not None:
                tr_invK_dK = jnp.einsum('ij,ijk->k', invK, dK)
                r_invK_dK_invK_r = jnp.einsum('i,ijk,j->k', invKr, dK, invKr)
                out['gradfwd'] += 1/2 * (tr_invK_dK - r_invK_dK_invK_r)
            if dr is not None:
                r_invK_dr = invKr @ dr
                out['gradfwd'] += r_invK_dr
        else:
            out['gradfwd'] = None

        if fisher:
            # = 1/2 tr(K⁻¹dKK⁻¹d'K)
            #   + dr'K⁻¹d'r
            # tr(K⁻¹dKK⁻¹d'K)_ij = tr(L'⁻¹L⁻¹dKL'⁻¹L⁻¹d'K)_ij =
            #                    = tr(L⁻¹dKL'⁻¹L⁻¹d'KL'⁻¹)_ij =
            #                    = (L⁻¹dKL'⁻¹)_kli (L⁻¹dKL'⁻¹)_klj
            # (L⁻¹dKL'⁻¹)_ijk = L⁻¹_il dK_lmk L'⁻¹_mj =
            #                 = L⁻¹_il L⁻¹_jm dK_lmk
            # (dr'K⁻¹d'r)_kq = dr'_k L'⁻¹L⁻¹dr_q =
            #                = (L⁻¹dr_k)_i (L⁻¹dr_q)_i
            out['fisher'] = 0
            if dK is not None:
                invL_dK = solve_triangular_batched(L,
                    jnp.moveaxis(dK, 2, 0),
                    lower=True) # kim: L⁻¹_il dK_lmk
                invL_dK_invL = solve_triangular_batched(L,
                    jnp.swapaxes(invL_dK, 1, 2),
                    lower=True) # kji: L⁻¹_jm (L⁻¹_il dK_lmk)
                tr_invK_dK_invK_dK = jnp.einsum('kij,qij->kq', invL_dK_invL, invL_dK_invL)
                out['fisher'] += 1/2 * tr_invK_dK_invK_dK
            if dr is not None:
                invLdr = jlinalg.solve_triangular(L, dr, lower=True)
                dr_invK_dr = invLdr.T @ invLdr
                out['fisher'] += dr_invK_dr
        else:
            out['fisher'] = None

        if fishvec:
            # = 1/2 tr(K⁻¹dKK⁻¹d'K) v
            #   + dr'K⁻¹d'r v
            # tr(K⁻¹dKK⁻¹d'K) v = K_vjp(K⁻¹K_jvp(v)K⁻¹) =
            #                   = K_vjp(L'⁻¹L⁻¹ K_jvp(v) L'⁻¹L⁻¹)
            # dr'K⁻¹d'r v = dr'K⁻¹dr_jvp(v) =
            #             = dr_vjp(K⁻¹dr_jvp(v)) =
            #             = dr_vjp(L'⁻¹L⁻¹ dr_jvp(v))
            out['fishvec'] = 0
            if not (dK_jvp_vec is None and dK_vjp is None):
                invL_dKv = jlinalg.solve_triangular(L, dK_jvp_vec, lower=True)
                invK_dKv = jlinalg.solve_triangular(L.T, invL_dKv, lower=False)
                invL_dKv_invK = jlinalg.solve_triangular(L, invK_dKv.T, lower=True)
                invK_dKv_invK = jlinalg.solve_triangular(L.T, invL_dKv_invK, lower=False)
                tr_invK_dK_invK_dK_v = dK_vjp(invK_dKv_invK)
                out['fishvec'] += 1/2 * tr_invK_dK_invK_dK_v
            if not (dr_jvp_vec is None and dr_vjp is None):
                invL_drv = jlinalg.solve_triangular(L, dr_jvp_vec, lower=True)
                invK_drv = jlinalg.solve_triangular(L.T, invL_drv, lower=False)
                dr_invK_drv_v = dr_vjp(invK_drv)
                out['fishvec'] += dr_invK_drv_v
        else:
            out['fishvec'] = None

        return tuple(out.values())

    @classmethod
    def make_derivs(cls,
        K_fun, r_fun, primal,
        *,
        args=(),
        kw={},
        vec=None,
        value=False,
        gradrev=False,
        gradfwd=False,
        fisher=False,
        fishvec=False,
    ):
        """
        Prepares arguments for `minus_log_normal_density`.

        Parameters
        ----------
        K_fun, r_fun : callable
            Functions with signature ``f(primal, *args, **kw)`` that produce the
            `K` init argument and the `r` `minus_log_normal_density` argument.
        primal : 1d array
            The first argument to `K_fun` and `r_fun`.
        args : tuple
            Additional positional arguments to `K_fun` and `r_fun`.
        kw : dict
            Keyword arguments to `K_fun` and `r_fun`.
        vec : 1d array
            A tangent vector to compute the jacobian-vector products.
        value, gradrev, gradfwd, fisher, fishvec : bool
            Arguments to `minus_log_normal_density`, used to determine which
            derivatives are needed.

        Returns
        -------
        K : 2d array
            Output of `K_fun`.
        r : 1d array
            Output of `r_fun`.
        out : dict
            Dictionary with derivative arguments to `minus_log_normal_density`.
        """

        partial = lambda f: lambda x: f(x, *args, **kw)
        K_fun = partial(K_fun)
        r_fun = partial(r_fun)

        out = {}

        if gradrev or fishvec:
            K, dK_vjp = jax.vjp(K_fun, primal)
            r, dr_vjp = jax.vjp(r_fun, primal)
            out['dK_vjp'] = lambda x: dK_vjp(x)[0]
            out['dr_vjp'] = lambda x: dr_vjp(x)[0]
        else:
            K = K_fun(primal)
            r = r_fun(primal)
        if fishvec:
            _, out['dK_jvp_vec'] = jax.jvp(K_fun, (primal,), (vec,))
            _, out['dr_jvp_vec'] = jax.jvp(r_fun, (primal,), (vec,))
        if gradfwd or fisher:
            out['dK'] = jax.jacfwd(K_fun)(primal)
            out['dr'] = jax.jacfwd(r_fun)(primal)

        return K, r, out

    @property
    def n(self):
        return len(self._L)

    m = n



# lsqfitgp/_gvarext/_ufunc.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools
import string

import jax
import gvar
from jax import numpy as jnp

from .. import _signature

from ._jacobian import jacobian, from_jacobian
from ._tabulate import tabulate_together

def gvar_gufunc(func, *, signature=None):
    """

    Wraps a jax-traceable generalized ufunc with one argument to support gvars.

    Parameters
    ----------
    func : callable
        A function from one array to one array. It must be a generalized ufunc,
        and differentiable one time with `jax`.
    signature : str, optional
        The signature of the generalized ufunc. If not specified, it is assumed
        to be scalar to scalar (normal ufunc).

    Returns
    -------
    decorated_func : callable
        A function that, in addition to numerical arrays, accepts gvars and
        returns gvars.

    See also
    --------
    numpy.vectorize

    """

    # parse signature
    if signature is None:
        signature = '()->()'
    sig = _signature.Signature(signature)
    inp, = sig.incores
    out, = sig.outcores
    jac_sig = _signature.Signature.from_tuples([inp], [out + inp])
    
    # make jacobian function
    deriv = jnp.vectorize(jax.jacfwd(func), signature=jac_sig.signature)

    # get indices for summation
    ninp = len(inp)
    nout = len(out)
    head_indices = '...'
    out_indices = string.ascii_letters[:nout]
    in_indices = string.ascii_letters[nout:nout + ninp]
    gvar_indices = string.ascii_letters[nout + ninp]

    # make summation formula
    jac_indices = head_indices + out_indices + in_indices
    in_jac_indices = head_indices + in_indices + gvar_indices
    out_indices = head_indices + out_indices + gvar_indices
    formula = f'{jac_indices},{in_jac_indices}->{out_indices}'

    def gvar_function(x):

        # unpack the gvars
        in_mean = gvar.mean(x)
        in_jac, indices = jacobian(x)

        # apply function
        out_mean = func(in_mean)
        jac = deriv(in_mean)

        # check shapes match
        head_ndim = jac.ndim - nout - ninp
        assert jac.shape[:head_ndim] == in_jac.shape[:in_jac.ndim - 1 - ninp]

        # contract
        out_jac = jnp.einsum(formula, jac, in_jac)

        # pack output
        return from_jacobian(out_mean, out_jac, indices)

    @functools.wraps(func)
    def decorated_func(x):
        if isinstance(x, gvar.GVar):
            out = gvar_function(x)
            if not out.ndim:
                out = out.item()
            return out
        elif getattr(x, 'dtype', None) == object:
            return gvar_function(x)
        else:
            return func(x)

    return decorated_func

    # TODO add more than one argument or output. Possibly without taking
    # derivatives when it's not a gvar, i.e., merge the wrappers and cycle over
    # args. Also implement excluded => note that jnp.vectorize only supports
    # positional arguments, excluded takes in only indices, not names
